"""Tools for fitting Bingham distributions to orientation distribution
functions (ODF), as described in :footcite:p:`Riffert2014`. The resulting
distributions can further be used to compute ODF-lobe-specific measures such as
the fiber density (FD) and fiber spread (FS) :footcite:p:`Riffert2014` and the
orientation dispersion index (ODI) :footcite:p:`NetoHenriques2018`.

References
----------
.. footbibliography::
"""

import numpy as np

from dipy.core.ndindex import ndindex
from dipy.core.onetime import auto_attr
from dipy.core.sphere import unit_icosahedron
from dipy.direction import peak_directions
from dipy.reconst.shm import calculate_max_order, sh_to_sf


def _bingham_fit_peak(sf, peak, sphere, max_search_angle):
    """
    Fit Bingham function on the ODF lobe aligned with peak.

    Parameters
    ----------
    sf: 1d ndarray
        The odf spherical function (sf) evaluated on the vertices
        of `sphere`.
    peak: ndarray (3, 1)
        The peak direction of the lobe to fit.
    sphere: `Sphere` class instance
        The Sphere providing the odf's discrete directions
    max_angle: float
        The maximum angle in degrees of the neighbourhood around the
        peak to consider for fitting.

    Returns
    -------
    f0: float
        Maximum amplitude of the ODF peak.
    k1: tuple of floats
        Concentration parameter of Bingham's major axis k1.
    k2: tuple of floats
        Concentration parameter of Bingham's minor axis k2.
    mu0: ndarray (3,) of floats
        Bingham's main axis.
    mu1: ndarray (3,) of floats
        Bingham's major concentration axis.
    mu2: ndarray (3,) of floats
        Bingham's minor concentration axis.
    """
    # abs for twice the number of pts to fit
    dot_prod = np.abs(sphere.vertices.dot(peak))
    min_dot = np.cos(np.deg2rad(max_search_angle))

    # [p] are the selected ODF vertices (N, 3) around the peak of the lobe
    # within max_angle
    p = sphere.vertices[dot_prod > min_dot]
    # [v] are the selected ODF amplitudes (N, 1) around the peak of the lobe
    # within max_angle
    v = sf[dot_prod > min_dot]

    # Test that the surface along peak direction contains
    # at least 3 non-zero directions
    if np.count_nonzero(v) < 3:
        return 0, 0.0, 0.0, np.zeros(3), np.zeros(3)

    x, y, z = (p[:, 0], p[:, 1], p[:, 2])

    # Create an orientation matrix (T) to approximate mu0, mu1 and mu2
    T = np.array(
        [
            [x**2 * v, x * y * v, x * z * v],
            [x * y * v, y**2 * v, y * z * v],
            [x * z * v, y * z * v, z**2 * v],
        ]
    )

    T = np.sum(T, axis=-1) / np.sum(v)

    # eigh better than eig. T will always be symmetric, eigh is faster.
    evals, evecs = np.linalg.eigh(T)

    # Not ordering the evals, eigh orders by default.
    mu0 = evecs[:, 2]
    mu1 = evecs[:, 1]
    mu2 = evecs[:, 0]
    f0 = v.max()  # Maximum amplitude of the ODF peak

    # If no real fit is possible, return null
    if np.iscomplex(mu1).any() or np.iscomplex(mu2).any():
        return 0, 0.0, 0.0, np.zeros(3), np.zeros(3)

    # Calculating the A matrix
    A = np.zeros((len(v), 2), dtype=float)  # (N, 2)
    A[:, 0] = p.dot(mu1) ** 2
    A[:, 1] = p.dot(mu2) ** 2

    # Test that AT.A is invertible for pseudo-inverse
    ATA = A.T.dot(A)
    if np.linalg.matrix_rank(ATA) != ATA.shape[0]:
        return 0, 0.0, 0.0, np.zeros(3), np.zeros(3)

    # Calculating the Beta matrix
    B = np.zeros_like(v)
    B[v > 0] = np.log(v[v > 0] / f0)  # (N, 1)

    # Calculating the Kappas
    k = np.abs(np.linalg.pinv(ATA).dot(A.T).dot(B))
    k1 = k[0]
    k2 = k[1]
    if k1 > k2:
        k1, k2 = k2, k1
        mu1, mu2 = mu2, mu1

    return f0, k1, k2, mu0, mu1, mu2


def _single_sf_to_bingham(
    odf, sphere, max_search_angle, *, npeaks=5, min_sep_angle=60, rel_th=0.1
):
    """
    Fit a Bingham distribution onto each principal ODF lobe.

    Parameters
    ----------
    odf: 1d ndarray
        The ODF function evaluated on the vertices of `sphere`
    sphere: `Sphere` class instance
        The Sphere providing the odf's discrete directions
    max_search_angle: float.
        Maximum angle between a peak and its neighbour directions
        for fitting the Bingham distribution. Although they suggest 6 degrees
        in :footcite:p:`Riffert2014`, tests show that a value around 45 degrees
        is more stable.
    npeak: int, optional
        Maximum number of peaks found (default 5 peaks).
    min_sep_angle: float, optional
        Minimum separation angle between two peaks for peak extraction.
    rel_th: float, optional
        Relative threshold used for peak extraction.

    Returns
    -------
    fits: list of tuples
        Bingham distribution parameters for each ODF peak.
    n: float
        Number of maximum peaks for the input ODF.

    Notes
    -----
    Lobes are first found by performing a peak extraction on the input ODF and
    Bingham distributions are then fitted around each of the extracted peaks
    using the method described in :footcite:p:`Riffert2014`.

    References
    ----------
    .. footbibliography::
    """
    # extract all maxima on the ODF
    dirs, vals, inds = peak_directions(
        odf, sphere, relative_peak_threshold=rel_th, min_separation_angle=min_sep_angle
    )

    # n becomes the new limit of peaks and sets a maximum of peaks in case
    # the ODF has more than npeaks.
    n = min(npeaks, vals.shape[0])

    # Calculate dispersion on all and each of the peaks up to 'n'
    if vals.shape[0] != 0:
        fits = []
        for i in range(n):
            fit = _bingham_fit_peak(odf, dirs[i], sphere, max_search_angle)
            fits.append(fit)

    return fits, n


def _single_bingham_to_sf(f0, k1, k2, major_axis, minor_axis, vertices):
    """
    Sample a Bingham function on the directions described by `vertices`.
    The function assumes that `vertices` are unit length and no checks
    are performed to validate that this is the case.

    Parameters
    ----------
    f0: float
        Maximum value of the Bingham function.
    k1: float
        Concentration along major axis.
    k2: float
        Concentration along minor axis.
    major_axis: ndarray (3)
        Direction of major axis
    minor_axis: ndarray (3)
        Direction of minor axis
    vertices: ndarray (N, 3)
        Unit sphere directions along which the distribution
        is evaluated.

    Returns
    -------
    fn : array (N,)
        Sampled Bingham function values at each point on directions.

    Notes
    -----
    Refer to method `bingham_to_odf` for the definition of
    the Bingham distribution.
    """
    x = -k1 * vertices.dot(major_axis) ** 2 - k2 * vertices.dot(minor_axis) ** 2
    fn = f0 * np.exp(x)

    return fn.T


def bingham_to_sf(bingham_params, sphere, *, mask=None):
    """
    Reconstruct ODFs from fitted Bingham parameters on multiple voxels.

    Parameters
    ----------
    bingham_params : ndarray (...., nl, 12)
        ndarray containing the model parameters of Binghams fitted to ODFs in
        the following order:
        - Maximum value of the Bingham function (f0, index 0);
        - concentration parameters k1 and k2 (indexes 1 and 2);
        - elements of Bingham's main direction (indexes 3-5);
        - elements of Bingham's dispersion major axis (indexes 6-8);
        - elements of Bingham's dispersion minor axis (indexes 9-11).
    sphere: `Sphere` class instance
         The Sphere providing the odf's discrete directions
    mask: ndarray, optional
        Map marking the coordinates in the data that should be analyzed.
        Default (None) means all voxels in the volume will be analyzed.

    Returns
    -------
    ODF : ndarray (..., n_directions)
            The value of the odf on each point of `sphere`.
    """
    n_directions = sphere.vertices.shape[0]

    shape = bingham_params.shape[0:-2]
    if mask is None:
        mask = np.ones(shape)

    odf = np.zeros(shape + (n_directions,))
    for idx in ndindex(shape):
        if not mask[idx]:
            continue
        bpars = bingham_params[idx]
        f0s = bpars[..., 0]
        npeaks = np.sum(f0s > 0)

        this_odf = 0

        for li in range(npeaks):
            f0 = f0s[li]
            k1 = bpars[li, 1]
            k2 = bpars[li, 2]
            mu1 = bpars[li, 6:9]
            mu2 = bpars[li, 9:12]

            this_odf += _single_bingham_to_sf(f0, k1, k2, mu1, mu2, sphere.vertices)
        odf[idx] = this_odf

    return odf


def bingham_fiber_density(bingham_params, *, subdivide=5, mask=None):
    """
    Compute fiber density for each lobe for a given Bingham ODF.
    Measured in the unit 1/mm^3.

    Parameters
    ----------
    bingham_params : ndarray (...., nl, 12)
        ndarray containing the model parameters of Bingham's fitted to ODFs in
        the following order:
        - Maximum value of the Bingham function (f0, index 0);
        - Concentration parameters k1 and k2 (indexes 1 and 2);
        - Elements of Bingham's main direction (indexes 3-5);
        - Elements of Bingham's dispersion major axis (indexes 6-8);
        - Elements of Bingham's dispersion minor axis (indexes 9-11).
    subdivide: int >= 0, optional
        Number of times the unit icosahedron used for integration
        should be subdivided. The higher this value the more precise the
        approximation will be, at the cost of longer execution times. The
        default results in a sphere of 10242 points.
    mask: ndarray, optional
        Map marking the coordinates in the data that should be analyzed.
        Default (None) means all voxels in the volume will be analyzed.

    Returns
    -------
    fd: ndarray (...., nl)
        Fiber density for each Bingham function.

    Notes
    -----
    Fiber density (fd) is given by the surface integral of the
    Bingham function :footcite:p:`Riffert2014`.

    References
    ----------
    .. footbibliography::
    """
    sphere = unit_icosahedron.subdivide(n=subdivide)

    # directions for evaluating the integral
    u = sphere.vertices

    # area of a single surface element
    dA = 4.0 * np.pi / len(u)

    shape = bingham_params.shape[0:-2]
    if mask is None:
        mask = np.ones(shape)

    # loop integral calculation for each image voxel
    fd = np.zeros(bingham_params.shape[0:-1])
    for idx in ndindex(shape):
        if not mask[idx]:
            continue

        bpars = bingham_params[idx]
        f0s = bpars[..., 0]
        npeaks = np.sum(f0s > 0)

        for li in range(npeaks):
            f0 = f0s[li]
            k1 = bpars[li, 1]
            k2 = bpars[li, 2]
            mu1 = bpars[li, 6:9]
            mu2 = bpars[li, 9:12]

            bingham_eval = _single_bingham_to_sf(f0, k1, k2, mu1, mu2, u)
            fd[idx + (li,)] = np.sum(bingham_eval * dA)

    return fd


def bingham_fiber_spread(f0, fd):
    """
    Compute fiber spread for each lobe for a given Bingham volume.
    Measured in radians.

    Parameters
    ----------
    f0: ndarray
        Peak amplitude (f0) of each Bingham function.
    fd: ndarray
        Fiber density (fd) of each Bingham function.

    Returns
    -------
    fs: list of floats
        Fiber spread (fs) of each each Bingham function.

    Notes
    -----
    Fiber spread (fs) is defined as fs = fd/f0 and characterizes the spread of
    the lobe, i.e. the higher the fs, the wider the lobe :footcite:p:`Riffert2014`.

    References
    ----------
    .. footbibliography::
    """
    fs = np.zeros(f0.shape)
    fs[f0 > 0] = fd[f0 > 0] / f0[f0 > 0]

    return fs


def k2odi(k):
    r"""
    Convert the Bingham/Watson concentration parameter k to the orientation
    dispersion index (ODI).

    Parameters
    ----------
    k: ndarray
        Watson/Bingham concentration parameter

    Returns
    -------
    ODI: float or ndarray
        Orientation Dispersion Index

    Notes
    -----
    Orientation Dispersion Index for Watson/Bingham functions are defined as
    :footcite:p:`NetoHenriques2018`, :footcite:p:`Zhang2012`:

    .. math::

        ODI = \frac{2}{pi} \arctan{( \frac{1}{k})}

    References
    ----------
    .. footbibliography::
    """
    odi = np.zeros(k.shape)
    odi[k > 0] = 2 / np.pi * np.arctan(1 / k[k > 0])
    return odi


def odi2k(odi):
    r"""
    Convert the orientation dispersion index (ODI) to the Bingham/Watson
    concentration parameter k.

    Parameters
    ----------
    ODI: ndarray
        Orientation Dispersion Index

    Returns
    -------
    k: float or ndarray
        Watson/Bingham concentration parameter

    Notes
    -----
    Orientation Dispersion Index for Watson/Bingham functions are defined as
    :footcite:p:`NetoHenriques2018`, :footcite:p:`Zhang2012`:

    .. math::

        ODI = \frac{2}{pi} \arctan ( \frac{1}{k} )

    References
    ----------
    .. footbibliography::
    """
    k = np.zeros(odi.shape)
    k[odi > 0] = 1 / np.tan(np.pi / 2 * odi[odi > 0])
    return k


def _convert_bingham_pars(fits, npeaks):
    """
    Convert list of tuples output of the Bingham fit to ndarray.

    Parameters
    ----------
    fits : tuple
        Tuple of nl elements containing the Bingham function parameters
        in the following order:
        - Maximum value of the Bingham function (f0);
        - concentration parameters (k1 and k2);
        - elements of Bingham's main direction (mu0);
        - elements of Bingham's dispersion major axis (mu1);
        - and elements of Bingham's dispersion minor axis (mu2).
    npeaks: int
        Maximum number of fitted Bingham functions, by number of peaks.

    Returns
    -------
    bingham_params : ndarray (nl, 12)
        ndarray containing the model parameters of Bingham fitted to ODFs in
        the following order:
        - Maximum value of the Bingham function (f0, index 0);
        - concentration parameters k1 and k2 (indexes 1 and 2);
        - elements of Bingham's main direction (indexes 3-5);
        - elements of Bingham's dispersion major axis (indexes 6-8);
        - elements of Bingham's dispersion minor axis (indexes 9-11).
    """
    n = len(fits)
    bpars = np.zeros((npeaks, 12))
    for ln in range(n):
        bpars[ln, 0] = fits[ln][0]
        bpars[ln, 1] = fits[ln][1]
        bpars[ln, 2] = fits[ln][2]
        bpars[ln, 3:6] = fits[ln][3]
        bpars[ln, 6:9] = fits[ln][4]
        bpars[ln, 9:12] = fits[ln][5]
    return bpars


def weighted_voxel_metric(bmetric, bfd):
    """
    Compute density-weighted scalar maps for metrics of Bingham functions
    fitted to multiple ODF lobes. The metric is computed as the
    weighted average of a given metric across the multiple ODF lobes
    (weights are defined by the Bingham fiber density estimates).

    Parameters
    ----------
    bmetric: ndarray(..., nl)
        Any metric with values for nl ODF lobes.
    bfd: ndarray(..., nl)
        Bingham's fiber density estimates for the nl ODF lobes

    Returns
    -------
    wmetric: ndarray(...)
        Weight-averaged Bingham metric
    """
    return np.sum(bmetric * bfd, axis=-1) / np.sum(bfd, axis=-1)


class BinghamMetrics:
    """
    Class for Bingham Metrics.
    """

    def __init__(self, model_params):
        """Initialization of the Bingham Metrics Class.

        Parameters
        ----------
        model_params : ndarray (..., nl, 12)
            ndarray containing Bingham's model parameters fitted to ODFs
            in the following order:
            - Maximum value of the Bingham function (f0, index 0);
            - concentration parameters k1 and k2 (indexes 1 and 2);
            - elements of Bingham's main direction (indexes 3-5);
            - elements of Bingham's dispersion major axis (indexes 6-8);
            - elements of Bingham's dispersion minor axis (indexes 9-11).
        """
        self.model_params = model_params

        self.peak_dirs = model_params[..., 3:6]

    @auto_attr
    def amplitude_lobe(self):
        """Maximum Bingham Amplitude for each ODF lobe.
        Measured in the unit 1/mm^3*rad."""
        return self.model_params[..., 0]

    @auto_attr
    def kappa1_lobe(self):
        """Concentration parameter k1 for each ODF lobe."""
        return self.model_params[..., 1]

    @auto_attr
    def kappa2_lobe(self):
        """Concentration parameter k2 for each ODF lobe."""
        return self.model_params[..., 2]

    @auto_attr
    def kappa_total_lobe(self):
        """Overall concentration parameters for an ODF peak.

        The overall (combined) concentration parameters for each lobe is
        defined by equation 19 in :footcite:p:`Tariq2016` as

        .. math::
            k_{total} = sqrt{(k_1 * k_2)}

        References
        ----------
        .. footbibliography::
        """
        return np.sqrt(self.kappa1_lobe * self.kappa2_lobe)

    @auto_attr
    def odi1_lobe(self):
        """Orientation Dispersion index 1 computed for each ODF lobe from
        concentration parameter kappa1."""
        return k2odi(self.kappa1_lobe)

    @auto_attr
    def odi2_lobe(self):
        """Orientation Dispersion index 2 computed for each ODF lobe from
        concentration parameter kappa2."""
        return k2odi(self.kappa2_lobe)

    @auto_attr
    def odi_total_lobe(self):
        """Overall Orientation Dispersion Index (ODI) for an ODF lobe.

        Overall Orientation Dispersion Index (ODI) computed for an
        ODF lobe from the overall concentration parameter (k_total).
        Defined by equation 20 in :footcite:p:`Tariq2016`.

        References
        ----------
        .. footbibliography::
        """
        return k2odi(self.kappa_total_lobe)

    @auto_attr
    def fd_lobe(self):
        """Fiber Density computed as the integral of the Bingham functions
        fitted for each ODF lobe."""
        return bingham_fiber_density(self.model_params)

    @auto_attr
    def odi1_voxel(self):
        """Voxel Orientation Dispersion Index 1 (weighted average of odi1
        across all lobes where the weights are each lobe's fd estimate).
        """
        return weighted_voxel_metric(self.odi1_lobe, self.fd_lobe)

    @auto_attr
    def odi2_voxel(self):
        """Voxel Orientation Dispersion Index 2 (weighted average of odi2
        across all lobes where the weights are each lobe's fd estimate).
        """
        return weighted_voxel_metric(self.odi2_lobe, self.fd_lobe)

    @auto_attr
    def odi_total_voxel(self):
        """Voxel total Orientation Dispersion Index (weighted average of
        odf_total across all lobes where the weights are each lobe's
        fd estimate)."""
        return weighted_voxel_metric(self.odi_total_lobe, self.fd_lobe)

    @auto_attr
    def fd_voxel(self):
        """Voxel fiber density (sum of fd estimates of all ODF lobes)."""
        return np.sum(self.fd_lobe, axis=-1)

    @auto_attr
    def fs_lobe(self):
        """Fiber spread computed for each ODF lobe.

        Notes
        -----
        Fiber spread (fs) is defined as fs = fd/f0 and characterizes the
        spread of the lobe, i.e. the higher the fs, the wider the lobe
        :footcite:p:`Riffert2014`.

        References
        ----------
        .. footbibliography::
        """
        return bingham_fiber_spread(self.amplitude_lobe, self.fd_lobe)

    @auto_attr
    def fs_voxel(self):
        """Voxel fiber spread (weighted average of fiber spread across all
        lobes where the weights are each lobe's fd estimate)."""
        return weighted_voxel_metric(self.fs_lobe, self.fd_lobe)

    def odf(self, sphere):
        """Reconstruct ODFs from fitted Bingham parameters on multiple voxels.

        Parameters
        ----------
        sphere: `Sphere` class instance
            The Sphere providing the discrete directions for ODF
            reconstruction.

        Returns
        -------
        ODF : ndarray (..., n_directions)
            The value of the odf on each point of `sphere`.
        """
        mask = self.fd_voxel > 0
        return bingham_to_sf(self.model_params, sphere, mask=mask)


def sf_to_bingham(
    odf, sphere, max_search_angle, *, mask=None, npeaks=5, min_sep_angle=60, rel_th=0.1
):
    """
    Fit the Bingham function from an image volume of ODFs.

    Parameters
    ----------
    odf: ndarray (Nx, Ny, Nz, Ndirs)
        Orientation Distribution Function sampled on the vertices of a sphere.
    sphere: `Sphere` class instance
        The Sphere providing the odf's discrete directions.
    max_search_angle: float.
        Maximum angle between a peak and its neighbour directions
        for fitting the Bingham distribution.
    mask: ndarray, optional
        Map marking the coordinates in the data that should be analyzed.
    npeak: int, optional
        Maximum number of peaks found.
    min_sep_angle: float, optional
        Minimum separation angle between two peaks for peak extraction.
    rel_th: float, optional
        Relative threshold used for peak extraction.

    Returns
    -------
    BinghamMetrics: class instance
        Class instance containing metrics computed from Bingham functions
        fitted to ODF lobes.
    """
    shape = odf.shape[0:-1]

    if mask is None:
        mask = np.ones(shape)

    # Bingham parameters stored in an ndarray with shape:
    # (Nx, Ny, Nz, n_max_peak, 12).
    bpars = np.zeros(shape + (npeaks,) + (12,))

    for idx in ndindex(shape):
        if not mask[idx]:
            continue

        [fits, npeaks_final] = _single_sf_to_bingham(
            odf[idx],
            sphere,
            max_search_angle,
            npeaks=npeaks,
            min_sep_angle=min_sep_angle,
            rel_th=rel_th,
        )

        bpars[idx] = _convert_bingham_pars(fits, npeaks)

    return BinghamMetrics(bpars)


def sh_to_bingham(
    sh,
    sphere,
    max_search_angle,
    *,
    mask=None,
    sh_basis="descoteaux07",
    legacy=True,
    npeaks=5,
    min_sep_angle=60,
    rel_th=0.1,
):
    """
    Fit the Bingham function from an image volume of spherical harmonics (SH)
    representing ODFs.

    Parameters
    ----------
    sh : ndarray
        SH coefficients representing a spherical function.
    sphere : `Sphere` class instance
        The Sphere providing the odf's discrete directions.
    max_search_angle: float.
        Maximum angle between a peak and its neighbour directions
        for fitting the Bingham distribution.
    mask: ndarray, optional
        Map marking the coordinates in the data that should be analyzed.
    sh_basis: str, optional
        SH basis. Either `descoteaux07` or `tournier07`.
    legacy: bool, optional
        Use legacy SH basis definitions.
    npeak: int, optional
        Maximum number of peaks found.
    min_sep_angle: float, optional
        Minimum separation angle between two peaks for peak extraction.
    rel_th: float, optional
        Relative threshold used for peak extraction.

    Returns
    -------
    BinghamMetrics: class instance
        Class instance containing metrics computed from Bingham functions
        fitted to ODF lobes.
    """
    shape = sh.shape[0:-1]
    sh_order_max = calculate_max_order(sh.shape[-1])

    if mask is None:
        mask = np.ones(shape)

    # Bingham parameters saved in an ndarray with shape:
    # (Nx, Ny, Nz, n_max_peak, 12).
    bpars = np.zeros(shape + (npeaks,) + (12,))

    for idx in ndindex(shape):
        if not mask[idx]:
            continue

        odf = sh_to_sf(
            sh[idx],
            sphere,
            sh_order_max=sh_order_max,
            basis_type=sh_basis,
            legacy=legacy,
        )

        [fits, npeaks_final] = _single_sf_to_bingham(
            odf,
            sphere,
            max_search_angle,
            npeaks=npeaks,
            min_sep_angle=min_sep_angle,
            rel_th=rel_th,
        )

        bpars[idx] = _convert_bingham_pars(fits, npeaks)

    return BinghamMetrics(bpars)



from os.path import join as pjoin
from tempfile import TemporaryDirectory
import warnings

import numpy as np
import numpy.testing as npt

from dipy.data import get_fnames
from dipy.io.image import load_nifti, load_nifti_data, save_nifti
from dipy.reconst.shm import descoteaux07_legacy_msg
from dipy.workflows.reconst import ReconstDsiFlow


def test_reconst_dsi():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstDsiFlow)


def test_reconst_dsid():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstDsiFlow, remove_convolution=True)


def reconst_flow_core(flow, **kwargs):
    with TemporaryDirectory() as out_dir:
        data_path, bval_path, bvec_path = get_fnames(name="small_64D")
        volume, affine = load_nifti(data_path)
        mask = np.ones_like(volume[:, :, :, 0])
        mask_path = pjoin(out_dir, "tmp_mask.nii.gz")
        save_nifti(mask_path, mask.astype(np.uint8), affine)

        dsi_flow = ReconstDsiFlow()
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )
            dsi_flow.run(
                data_path,
                bval_path,
                bvec_path,
                mask_path,
                out_dir=out_dir,
                extract_pam_values=True,
                **kwargs,
            )

        peaks_dir_path = dsi_flow.last_generated_outputs["out_peaks_dir"]
        peaks_dir_data = load_nifti_data(peaks_dir_path)
        npt.assert_equal(peaks_dir_data.shape[-1], 15)
        npt.assert_equal(peaks_dir_data.shape[:-1], volume.shape[:-1])

        peaks_idx_path = dsi_flow.last_generated_outputs["out_peaks_indices"]
        peaks_idx_data = load_nifti_data(peaks_idx_path)
        npt.assert_equal(peaks_idx_data.shape[-1], 5)
        npt.assert_equal(peaks_idx_data.shape[:-1], volume.shape[:-1])

        peaks_vals_path = dsi_flow.last_generated_outputs["out_peaks_values"]
        peaks_vals_data = load_nifti_data(peaks_vals_path)
        npt.assert_equal(peaks_vals_data.shape[-1], 5)
        npt.assert_equal(peaks_vals_data.shape[:-1], volume.shape[:-1])

        gfa_path = dsi_flow.last_generated_outputs["out_gfa"]
        gfa_data = load_nifti_data(gfa_path)
        npt.assert_equal(gfa_data.shape, volume.shape[:-1])



"""
=====================================================
Calculation of Outliers with Cluster Confidence Index
=====================================================

This is an outlier scoring method that compares the pathways of each streamline
in a bundle (pairwise) and scores each streamline by how many other streamlines
have similar pathways. The details can be found in :footcite:p:`Jordan2018`.

"""

import matplotlib.pyplot as plt

from dipy.core.gradients import gradient_table
from dipy.data import default_sphere, get_fnames
from dipy.direction import peaks_from_model
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data
from dipy.reconst.shm import CsaOdfModel
from dipy.tracking import utils
from dipy.tracking.local_tracking import LocalTracking
from dipy.tracking.stopping_criterion import ThresholdStoppingCriterion
from dipy.tracking.streamline import Streamlines, cluster_confidence
from dipy.tracking.utils import length
from dipy.viz import actor, window

###############################################################################
# First, we need to generate some streamlines. For a more complete
# description of these steps, please refer to the CSA Probabilistic Tracking
# and the Visualization of ROI Surface Rendered with Streamlines Tutorials.

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")
label_fname = get_fnames(name="stanford_labels")

data, affine = load_nifti(hardi_fname)
labels = load_nifti_data(label_fname)
bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

white_matter = (labels == 1) | (labels == 2)
csa_model = CsaOdfModel(gtab, sh_order_max=6)
csa_peaks = peaks_from_model(
    csa_model,
    data,
    default_sphere,
    relative_peak_threshold=0.8,
    min_separation_angle=45,
    mask=white_matter,
)
stopping_criterion = ThresholdStoppingCriterion(csa_peaks.gfa, 0.25)

###############################################################################
# We will use a slice of the anatomically-based corpus callosum ROI as our
# seed mask to demonstrate the method.


# Make a corpus callosum seed mask for tracking
seed_mask = labels == 2
seeds = utils.seeds_from_mask(seed_mask, affine, density=[1, 1, 1])
# Make a streamline bundle model of the corpus callosum ROI connectivity
streamlines = LocalTracking(csa_peaks, stopping_criterion, seeds, affine, step_size=2)
streamlines = Streamlines(streamlines)

###############################################################################
# We do not want our results inflated by short streamlines, so we remove
# streamlines shorter than 40mm prior to calculating the CCI.

lengths = list(length(streamlines))
long_streamlines = Streamlines()
for i, sl in enumerate(streamlines):
    if lengths[i] > 40:
        long_streamlines.append(sl)

###############################################################################
# Now we calculate the Cluster Confidence Index using the corpus callosum
# streamline bundle and visualize them.

cci = cluster_confidence(long_streamlines)

# Visualize the streamlines, colored by cci
scene = window.Scene()

hue = [0.5, 1]
saturation = [0.0, 1.0]

lut_cmap = actor.colormap_lookup_table(
    scale_range=(cci.min(), cci.max() / 4), hue_range=hue, saturation_range=saturation
)

bar3 = actor.scalar_bar(lookup_table=lut_cmap)
scene.add(bar3)

stream_actor = actor.line(
    long_streamlines, colors=cci, linewidth=0.1, lookup_colormap=lut_cmap
)
scene.add(stream_actor)

###############################################################################
# If you set interactive to True (below), the scene will pop up in an
# interactive window.

interactive = False
if interactive:
    window.show(scene)
window.record(scene=scene, out_path="cci_streamlines.png", size=(800, 800))

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Cluster Confidence Index of corpus callosum dataset.
#
#
# If you think of each streamline as a sample of a potential pathway through a
# complex landscape of white matter anatomy probed via water diffusion,
# intuitively we have more confidence that pathways represented by many samples
# (streamlines) reflect a more stable representation of the underlying
# phenomenon we are trying to model (anatomical landscape) than do lone
# samples.
#
# The CCI provides a voting system where by each streamline (within a set
# tolerance) gets to vote on how much support it lends to. Outlier pathways
# score relatively low on CCI, since they do not have many streamlines voting
# for them. These outliers can be removed by thresholding on the CCI metric.

fig, ax = plt.subplots(1)
ax.hist(cci, bins=100, histtype="step")
ax.set_xlabel("CCI")
ax.set_ylabel("# streamlines")
fig.savefig("cci_histogram.png")


###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Histogram of Cluster Confidence Index values.
#
#
# Now we threshold the CCI, defining outliers as streamlines that score
# below 1.

keep_streamlines = Streamlines()
for i, sl in enumerate(long_streamlines):
    if cci[i] >= 1:
        keep_streamlines.append(sl)

# Visualize the streamlines we kept
scene = window.Scene()

keep_streamlines_actor = actor.line(keep_streamlines, linewidth=0.1)

scene.add(keep_streamlines_actor)

interactive = False
if interactive:
    window.show(scene)
window.record(scene=scene, out_path="filtered_cci_streamlines.png", size=(800, 800))

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Outliers, defined as streamlines scoring CCI < 1, were excluded.
#
#
# References
# ----------
#
# .. footbibliography::
#



cimport numpy as np

cdef class PmfGen:
    cdef:
        double[:] pmf
        double[:, :, :, :] data
        double[:, :] vertices
        object sphere

    cdef double* get_pmf_c(self, double* point, double* out) noexcept nogil
    cdef int find_closest(self, double* xyz) noexcept nogil
    cdef double get_pmf_value_c(self, double* point, double* xyz) noexcept nogil
    pass


cdef class SimplePmfGen(PmfGen):
    pass


cdef class SHCoeffPmfGen(PmfGen):
    cdef:
        double[:, :] B
        double[:] coeff
    pass



import numpy as np
import numpy.testing as npt

import dipy.utils.parallel as para


def power_it(num, n=2):
    # We define a function of the right form for parallelization
    return num**n


ENGINES = ["serial"]
if para.has_dask:
    ENGINES.append("dask")
if para.has_joblib:
    ENGINES.append("joblib")
if para.has_ray:
    ENGINES.append("ray")


def test_paramap():
    my_array = np.arange(100).reshape(10, 10)
    my_list = list(my_array.ravel())
    for engine in ENGINES:
        for backend in ["threading", "multiprocessing"]:
            npt.assert_array_equal(
                para.paramap(
                    power_it,
                    my_list,
                    engine=engine,
                    backend=backend,
                    out_shape=my_array.shape,
                ),
                power_it(my_array),
            )

            # If it's not reshaped, the first item should be the item 0, 0:
            npt.assert_equal(
                para.paramap(power_it, my_list, engine=engine, backend=backend)[0],
                power_it(my_array[0, 0]),
            )


def test_paramap_sequence_kwargs():
    my_array = np.arange(10)
    kwargs_sequence = [{"n": i} for i in range(len(my_array))]
    my_list = list(my_array.ravel())
    for engine in ENGINES:
        for backend in ["threading", "multiprocessing"]:
            npt.assert_array_equal(
                para.paramap(
                    power_it,
                    my_list,
                    engine=engine,
                    backend=backend,
                    out_shape=my_array.shape,
                    func_kwargs=kwargs_sequence,
                ),
                [
                    power_it(ii, **kwargs)
                    for ii, kwargs in zip(my_array, kwargs_sequence)
                ],
            )



"""
===============================================
Reconstruct with Generalized Q-Sampling Imaging
===============================================

We show how to apply Generalized Q-Sampling Imaging :footcite:p:`Yeh2010`
to diffusion MRI datasets. You can think of GQI as an analytical version of
DSI orientation distribution function (ODF) (Garyfallidis, PhD thesis, 2012).

First import the necessary modules:
"""

import numpy as np

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames, get_sphere
from dipy.direction import peaks_from_model
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.reconst.gqi import GeneralizedQSamplingModel

###############################################################################
# Download and get the data filenames for this tutorial.

fraw, fbval, fbvec = get_fnames(name="taiwan_ntu_dsi")

###############################################################################
# img contains a nibabel Nifti1Image object (data) and gtab contains a
# ``GradientTable`` object (gradient information e.g. b-values). For example
# to read the b-values it is possible to write::
#
#    print(gtab.bvals)
#
# Load the raw diffusion data and the affine.

data, affine, voxel_size = load_nifti(fraw, return_voxsize=True)
bvals, bvecs = read_bvals_bvecs(fbval, fbvec)
bvecs[1:] = bvecs[1:] / np.sqrt(np.sum(bvecs[1:] * bvecs[1:], axis=1))[:, None]
gtab = gradient_table(bvals, bvecs=bvecs)
print(f"data.shape {data.shape}")

###############################################################################
# This dataset has anisotropic voxel sizes, therefore reslicing is necessary.
#
# Instantiate the model and apply it to the data.

gqmodel = GeneralizedQSamplingModel(gtab, sampling_length=3)

###############################################################################
# The parameter ``sampling_length`` is used here to
#
# Lets just use one slice only from the data.

dataslice = data[:, :, data.shape[2] // 2]

mask = dataslice[..., 0] > 50

gqfit = gqmodel.fit(dataslice, mask=mask)

###############################################################################
# Load an ODF reconstruction sphere

sphere = get_sphere(name="repulsion724")

###############################################################################
# Calculate the ODFs with this specific sphere

ODF = gqfit.odf(sphere)

print(f"ODF.shape {ODF.shape}")

###############################################################################
# Using ``peaks_from_model`` we can find the main peaks of the ODFs and other
# properties.

gqpeaks = peaks_from_model(
    model=gqmodel,
    data=dataslice,
    sphere=sphere,
    relative_peak_threshold=0.5,
    min_separation_angle=25,
    mask=mask,
    return_odf=False,
    normalize_peaks=True,
)

gqpeak_values = gqpeaks.peak_values

###############################################################################
# ``gqpeak_indices`` show which sphere points have the maximum values.

gqpeak_indices = gqpeaks.peak_indices

###############################################################################
# It is also possible to calculate GFA.

GFA = gqpeaks.gfa

print(f"GFA.shape {GFA.shape}")

###############################################################################
# With parameter ``return_odf=True`` we can obtain the ODF using
# ``gqpeaks.ODF``

gqpeaks = peaks_from_model(
    model=gqmodel,
    data=dataslice,
    sphere=sphere,
    relative_peak_threshold=0.5,
    min_separation_angle=25,
    mask=mask,
    return_odf=True,
    normalize_peaks=True,
)

###############################################################################
# This ODF will be of course identical to the ODF calculated above as long as
# the same data and mask are used.

print(np.sum(gqpeaks.odf != ODF) == 0)

###############################################################################
# The advantage of using ``peaks_from_model`` is that it calculates the ODF
# only once and saves it or deletes if it is not necessary to keep.
#
#
# References
# ----------
#
# .. footbibliography::
#



import numpy as np
cimport numpy as cnp
cimport cython

from safe_openmp cimport have_openmp
from cython.parallel import parallel, prange, threadid

from scipy.spatial import KDTree
from scipy.interpolate import interp1d

from dipy.utils.omp import determine_num_threads
from dipy.utils.omp cimport set_num_threads, restore_default_num_threads

cdef class FBCMeasures:

    cdef int [:] streamline_length
    cdef double [:, :, :] streamline_points
    cdef double [:, :] streamlines_lfbc
    cdef double [:] streamlines_rfbc

    def __init__(self,
                 streamlines,
                 kernel,
                 min_fiberlength=10,
                 max_windowsize=7,
                 num_threads=None,
                 verbose=False):
        """ Compute the fiber to bundle coherence measures for a set of
        streamlines.

        See :footcite:p`Meesters2016b` and :footcite:p:`Portegies2015b` for
        further details about the method.

        Parameters
        ----------
        streamlines : list
            A collection of streamlines, each n by 3, with n being the number of
            nodes in the fiber.
        kernel : Kernel object
            A diffusion kernel object created from EnhancementKernel.
        min_fiberlength : int
            Fibers with fewer points than minimum_length are excluded from FBC
            computation.
        max_windowsize : int
            The maximal window size used to calculate the average LFBC region
        num_threads : int, optional
            Number of threads to be used for OpenMP parallelization. If None
            (default) the value of OMP_NUM_THREADS environment variable is used
            if it is set, otherwise all available threads are used. If < 0 the
            maximal number of threads minus |num_threads + 1| is used (enter -1
            to use as many threads as possible). 0 raises an error.
        verbose : boolean
            Enable verbose mode.

        References
        ----------
        .. footbibliography::
        """
        self.compute(streamlines,
                 kernel,
                 min_fiberlength=min_fiberlength,
                 max_windowsize=max_windowsize,
                 num_threads=num_threads,
                 verbose=verbose)

    def get_points_rfbc_thresholded(self, threshold, emphasis=.5, verbose=False):
        """ Set a threshold on the RFBC to remove spurious fibers.

        Parameters
        ----------
        threshold : float
            The threshold to set on the RFBC, should be within 0 and 1.
        emphasis : float
            Enhances the coloring of the fibers by LFBC. Increasing emphasis
            will stress spurious fibers by logarithmic weighting.
        verbose : boolean
            Prints info about the found RFBC for the set of fibers such as
            median, mean, min and max values.

        Returns
        -------
        output : tuple with 3 elements
            The output contains:
            1) a collection of streamlines, each n by 3, with n
            being the number of nodes in the fiber that remain after filtering
            2) the r,g,b values of the local fiber to bundle coherence (LFBC)
            3) the relative fiber to bundle coherence (RFBC)
        """
        if verbose:
            print("median RFBC: " + str(np.median(self.streamlines_rfbc)))
            print("mean RFBC: " + str(np.mean(self.streamlines_rfbc)))
            print("min RFBC: " + str(np.min(self.streamlines_rfbc)))
            print("max RFBC: " + str(np.max(self.streamlines_rfbc)))

        # logarithmic transform of color values to emphasize spurious fibers
        minval = np.nanmin(self.streamlines_lfbc)
        maxval = np.nanmax(self.streamlines_lfbc)
        lfbc_log = np.log((self.streamlines_lfbc - minval) /
                          (maxval - minval + 10e-10) + emphasis)
        minval = np.nanmin(lfbc_log)
        maxval = np.nanmax(lfbc_log)
        lfbc_log = (lfbc_log - minval) / (maxval - minval)

        # define color interpolation functions
        x = np.linspace(0, 1, num=4, endpoint=True)
        r = np.array([1, 1, 0, 0])
        g = np.array([1, 0, 0, 1])
        b = np.array([0, 0, 1, 1])
        fr = interp1d(x, r, bounds_error=False, fill_value=0)
        fg = interp1d(x, g, bounds_error=False, fill_value=0)
        fb = interp1d(x, b, bounds_error=False, fill_value=0)

        # select fibers above the RFBC threshold
        streamline_out = []
        color_out = []
        rfbc_out = []
        for i in range(self.streamlines_rfbc.shape[0]):
            rfbc = self.streamlines_rfbc[i]
            lfbc = lfbc_log[i]
            if rfbc > threshold:
                fiber = np.array(self.streamline_points[i])
                fiber = fiber[0:self.streamline_length[i] - 1]
                streamline_out.append(fiber)

                rfbc_out.append(rfbc)

                lfbc = lfbc[0:self.streamline_length[i] - 1]
                lfbc_colors = np.transpose([fr(lfbc), fg(lfbc), fb(lfbc)])
                color_out.append(lfbc_colors.tolist())

        return streamline_out, color_out, rfbc_out

    @cython.wraparound(False)
    @cython.boundscheck(False)
    @cython.nonecheck(False)
    @cython.cdivision(True)
    cdef void compute(self,
                      py_streamlines,
                      kernel,
                      min_fiberlength,
                      max_windowsize,
                      num_threads=None,
                      verbose=False):
        """ Compute the fiber to bundle coherence measures for a set of
        streamlines.

        Parameters
        ----------
        py_streamlines : list
            A collection of streamlines, each n by 3, with n being the number of
            nodes in the fiber.
        kernel : Kernel object
            A diffusion kernel object created from EnhancementKernel.
        min_fiberlength : int
            Fibers with fewer points than minimum_length are excluded from FBC
            computation.
        max_windowsize : int
            The maximal window size used to calculate the average LFBC region
        num_threads : int, optional
            Number of threads to be used for OpenMP parallelization. If None
            (default) the value of OMP_NUM_THREADS environment variable is used
            if it is set, otherwise all available threads are used. If < 0 the
            maximal number of threads minus |num_threads + 1| is used (enter -1
            to use as many threads as possible). 0 raises an error.
        verbose : boolean
            Enable verbose mode.
        """
        cdef:
            cnp.npy_intp num_fibers, max_length, dim
            double [:, :, :] streamlines
            int [:] streamlines_length
            double [:, :, :] streamlines_tangent
            int [:, :] streamlines_nearestp
            double [:, :] streamline_scores
            cnp.npy_intp line_id = 0
            cnp.npy_intp point_id = 0
            cnp.npy_intp line_id2 = 0
            cnp.npy_intp point_id2 = 0
            cnp.npy_intp dims
            double score
            double [:] score_mp
            int [:] xd_mp, yd_mp, zd_mp
            cnp.npy_intp xd, yd, zd, N, hn
            double [:, :, :, :, ::1] lut
            cnp.npy_intp threads_to_use = -1

        threads_to_use = determine_num_threads(num_threads)
        set_num_threads(threads_to_use)

        # if the fibers are too short FBC measures cannot be applied,
        # remove these.
        streamlines_length = np.array([x.shape[0] for x in py_streamlines],
                                      dtype=np.int32)
        min_length = min(streamlines_length)
        if min_length < min_fiberlength:
            print("The minimum fiber length is 10 points. \
                    Shorter fibers were found and removed.")
            py_streamlines = [x for x in py_streamlines
                              if x.shape[0] >= min_fiberlength]
            streamlines_length = np.array([x.shape[0] for x in py_streamlines],
                                          dtype=np.int32)
        num_fibers = len(py_streamlines)
        self.streamline_length = streamlines_length
        max_length = max(streamlines_length)

        dim = 3

        # get lookup table info
        lut = kernel.get_lookup_table()
        N = lut.shape[2]
        hn = (N-1) / 2

        # prepare numpy arrays for speed
        streamlines = np.zeros((num_fibers, max_length, dim),
                               dtype=np.float64) * np.nan
        streamlines_tangents = np.zeros((num_fibers, max_length, dim),
                                        dtype=np.float64)
        streamlines_nearestp = np.zeros((num_fibers, max_length),
                                        dtype=np.int32)
        streamline_scores = np.zeros((num_fibers, max_length),
                                     dtype=np.float64) * np.nan

        # copy python streamlines into numpy array
        for line_id in range(num_fibers):
            for point_id in range(streamlines_length[line_id]):
                for dims in range(3):
                    streamlines[line_id, point_id, dims] = \
                        py_streamlines[line_id][point_id][dims]
        self.streamline_points = streamlines

        # compute tangents
        for line_id in range(num_fibers):
            for point_id in range(streamlines_length[line_id] - 1):
                tangent = np.subtract(streamlines[line_id, point_id + 1],
                                      streamlines[line_id, point_id])
                streamlines_tangents[line_id, point_id] = \
                    np.divide(tangent,
                              np.sqrt(np.dot(tangent, tangent)))

        # estimate which kernel LUT index corresponds to angles
        tree = KDTree(kernel.get_orientations())
        for line_id in range(num_fibers):
            for point_id in range(streamlines_length[line_id] - 1):
                streamlines_nearestp[line_id, point_id] = \
                    tree.query(streamlines[line_id, point_id])[1]

        # arrays for parallel computing
        score_mp = np.zeros(num_fibers)
        xd_mp = np.zeros(num_fibers, dtype=np.int32)
        yd_mp = np.zeros(num_fibers, dtype=np.int32)
        zd_mp = np.zeros(num_fibers, dtype=np.int32)

        if verbose:
            if have_openmp:
                print("Running in parallel!")
            else:
                print("No OpenMP...")

        # compute fiber LFBC measures
        with nogil:

            for line_id in prange(num_fibers, schedule='guided'):
                for point_id in range(streamlines_length[line_id] - 1):
                    score_mp[line_id] = 0.0
                    for line_id2 in range(num_fibers):

                        # skip lfbc computation with itself
                        if line_id == line_id2:
                            continue

                        for point_id2 in range(streamlines_length[line_id2] - 1):
                            # compute displacement
                            xd_mp[line_id] = \
                                int(streamlines[line_id, point_id, 0] -
                                    streamlines[line_id2, point_id2, 0] + 0.5)
                            yd_mp[line_id] = \
                                int(streamlines[line_id, point_id, 1] -
                                    streamlines[line_id2, point_id2, 1] + 0.5)
                            zd_mp[line_id] = \
                                int(streamlines[line_id, point_id, 2] -
                                    streamlines[line_id2, point_id2, 2] + 0.5)

                            # if position is outside the kernel bounds, skip
                            if xd_mp[line_id] > hn or -xd_mp[line_id] > hn or \
                               yd_mp[line_id] > hn or -yd_mp[line_id] > hn or \
                               zd_mp[line_id] > hn or -zd_mp[line_id] > hn:
                                continue

                            # grab kernel value from LUT
                            score_mp[line_id] += \
                                lut[streamlines_nearestp[line_id, point_id],
                                    streamlines_nearestp[line_id2, point_id2],
                                    hn+xd_mp[line_id],
                                    hn+yd_mp[line_id],
                                    hn+zd_mp[line_id]]  # ang_v, ang_r, x, y, z

                    streamline_scores[line_id, point_id] = score_mp[line_id]

        # Reset number of OpenMP cores to default
        if num_threads is not None:
            restore_default_num_threads()

        # Save LFBC as class member
        self.streamlines_lfbc = streamline_scores

        # compute RFBC for each fiber
        self.streamlines_rfbc = compute_rfbc(streamlines_length,
                                             streamline_scores,
                                             max_windowsize=max_windowsize)


def compute_rfbc(streamlines_length, streamline_scores, max_windowsize=7):
    """ Compute the relative fiber to bundle coherence (RFBC)

    Parameters
    ----------
    streamlines_length : 1D int array
        Contains the length of each streamline
    streamlines_scores : 2D double array
        Contains the local fiber to bundle coherence (LFBC) for each streamline
        element.
    max_windowsize : int
        The maximal window size used to calculate the average LFBC region

    Returns
    -------
    output: normalized lowest average LFBC region along the fiber
    """

    # finds the region of the fiber with maximal length of max_windowsize in
    # which the LFBC is the lowest
    int_length = min(np.amin(streamlines_length), max_windowsize)
    int_value = np.apply_along_axis(lambda x: min_moving_average(x[~np.isnan(x)],
                                    int_length), 1, streamline_scores)
    avg_total = np.mean(
                    np.apply_along_axis(
                        lambda x: np.mean(np.extract(x[~np.isnan(x)] >= 0,
                                          x[~np.isnan(x)])),
                        1, streamline_scores))
    if not avg_total == 0:
        return int_value / avg_total
    else:
        return int_value


def min_moving_average(a, n):
    """ Return the lowest cumulative sum for the score of a streamline segment

    Parameters
    ----------
    a : array
        Input array

    n : int
        Length of the segment

    Returns
    -------
    output: normalized lowest average LFBC region along the fiber
    """
    ret = np.cumsum(np.extract(a >= 0, a))
    ret[n:] = ret[n:] - ret[:-n]
    return np.amin(ret[n - 1:] / n)



from collections.abc import Sequence
import json
import multiprocessing
import shutil
import tempfile

import numpy as np
from tqdm.auto import tqdm

from dipy.testing.decorators import warning_for_keywords
from dipy.utils.optpkg import optional_package

ray, has_ray, _ = optional_package("ray")
joblib, has_joblib, _ = optional_package("joblib")
dask, has_dask, _ = optional_package("dask")


@warning_for_keywords()
def paramap(
    func,
    in_list,
    *,
    out_shape=None,
    n_jobs=-1,
    engine="ray",
    backend=None,
    func_args=None,
    clean_spill=True,
    func_kwargs=None,
    **kwargs,
):
    # FIXME: but several fitting functions return "extra" as well
    #        is this being handled properly, or not?
    #        perhaps the 'func' docs below are misleading?
    """
    Maps a function to a list of inputs in parallel.

    Parameters
    ----------
    func : callable
        The function to apply to each item in the array. Must have the form:
        ``func(arr, idx, *args, *kwargs)`` where arr is an ndarray and idx is
        an index into that array (a tuple). The Return of `func` needs to be
        one item (e.g. float, int) per input item.
    in_list : list
       A sequence of items each of which can be an input to ``func``.
    out_shape : tuple, optional
         The shape of the output array. If not specified, the output shape will
         be `(len(in_list),)`.
    n_jobs : integer, optional
        The number of jobs to perform in parallel.
        -1 (default) to use all but one cpu.
    engine : str, optional
        {"dask", "joblib", "ray", "serial"}
        The last one is useful for debugging -- runs the code without any
        parallelization.
    backend : str, optional
        What joblib or dask backend to use. For joblib, the default is "loky".
        For dask the default is "threading".
    clean_spill : bool, optional
        If True, clean up the spill directory after the computation is done.
        Only applies to "ray" engine. Otherwise has no effect.
    func_args : list, optional
        Positional arguments to `func`.
    func_kwargs : dict or sequence, optional
        Keyword arguments to `func` or sequence of keyword arguments
        to `func`: one item for each item in the input list.
    kwargs : dict, optional
        Additional arguments to pass to either joblib.Parallel
        or dask.compute, or ray.remote depending on the engine used.

    Returns
    -------
    ndarray of identical shape to `arr`

    """

    func_args = func_args or []
    func_kwargs = func_kwargs or {}
    # Check if the func_kwargs are a sequence:

    if isinstance(func_kwargs, Sequence):
        if len(func_kwargs) != len(in_list):
            raise ValueError(
                "The length of func_kwargs should be the same as the length of in_list"
            )
        func_kwargs_sequence = True
    else:
        func_kwargs_sequence = False

    if engine == "joblib":
        if not has_joblib:
            raise joblib()
        if backend is None:
            backend = "loky"
        pp = joblib.Parallel(n_jobs=n_jobs, backend=backend, **kwargs)
        dd = joblib.delayed(func)
        if func_kwargs_sequence:
            d_l = [dd(ii, *func_args, **fk) for ii, fk in zip(in_list, func_kwargs)]
        else:
            d_l = [dd(ii, *func_args, **func_kwargs) for ii in in_list]
        results = pp(tqdm(d_l))

    elif engine == "dask":
        if not has_dask:
            raise dask()
        if backend is None:
            backend = "threading"

        if n_jobs == -1:
            n_jobs = multiprocessing.cpu_count()
            n_jobs = n_jobs - 1

        def partial(func, *args, **keywords):
            def newfunc(in_arg):
                return func(in_arg, *args, **keywords)

            return newfunc

        delayed_func = dask.delayed(func)

        if func_kwargs_sequence:
            dd = [
                delayed_func(ii, *func_args, **fk)
                for ii, fk in zip(in_list, func_kwargs)
            ]
        else:
            pp = dask.delayed(partial(func, *func_args, **func_kwargs))
            dd = [pp(ii) for ii in in_list]

        if backend == "multiprocessing":
            results = dask.compute(*dd, scheduler="processes", workers=n_jobs, **kwargs)
        elif backend == "threading":
            results = dask.compute(*dd, scheduler="threads", workers=n_jobs, **kwargs)
        else:
            raise ValueError(f"{backend} is not a backend for dask")

    elif engine == "ray":
        if not has_ray:
            raise ray()

        if clean_spill:
            tmp_dir = tempfile.TemporaryDirectory()

            if not ray.is_initialized():
                ray.init(
                    _system_config={
                        "object_spilling_config": json.dumps(
                            {
                                "type": "filesystem",
                                "params": {"directory_path": tmp_dir.name},
                            },
                        )
                    },
                )

        func = ray.remote(func)
        if func_kwargs_sequence:
            results = ray.get(
                [
                    func.remote(ii, *func_args, **fk)
                    for ii, fk in zip(in_list, func_kwargs)
                ]
            )
        else:
            results = ray.get(
                [func.remote(ii, *func_args, **func_kwargs) for ii in in_list]
            )

        if clean_spill:
            shutil.rmtree(tmp_dir.name)

    elif engine == "serial":
        results = []
        if func_kwargs_sequence:
            for in_element, fk in zip(in_list, func_kwargs):
                results.append(func(in_element, *func_args, **fk))
        else:
            for in_element in in_list:
                results.append(func(in_element, *func_args, **func_kwargs))
    else:
        raise ValueError("%s is not a valid engine" % engine)

    if out_shape is not None:
        return np.array(results).reshape(out_shape)
    else:
        return results



import numpy as np
from numpy.testing import assert_, assert_array_almost_equal

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.reconst.dti import TensorModel
from dipy.sims.phantom import orbital_phantom
from dipy.testing.decorators import set_random_number_generator

fimg, fbvals, fbvecs = get_fnames(name="small_64D")
bvals, bvecs = read_bvals_bvecs(fbvals, fbvecs)
bvecs[np.isnan(bvecs)] = 0

gtab = gradient_table(bvals, bvecs=bvecs)


def f(t):
    """
    Helper function used to define a mapping time => xyz
    """
    x = np.linspace(-1, 1, len(t))
    y = np.linspace(-1, 1, len(t))
    z = np.linspace(-1, 1, len(t))
    return x, y, z


def test_phantom():
    N = 50

    vol = orbital_phantom(
        gtab=gtab,
        func=f,
        t=np.linspace(0, 2 * np.pi, N),
        datashape=(10, 10, 10, len(bvals)),
        origin=(5, 5, 5),
        scale=(3, 3, 3),
        angles=np.linspace(0, 2 * np.pi, 16),
        radii=np.linspace(0.2, 2, 6),
        S0=100,
    )

    m = TensorModel(gtab)
    t = m.fit(vol)
    FA = t.fa
    # print vol
    FA[np.isnan(FA)] = 0
    # 686 -> expected FA given diffusivities of [1500, 400, 400]
    l1, l2, l3 = 1500e-6, 400e-6, 400e-6
    expected_fa = (
        np.sqrt(0.5)
        * np.sqrt((l1 - l2) ** 2 + (l2 - l3) ** 2 + (l3 - l1) ** 2)
        / np.sqrt(l1**2 + l2**2 + l3**2)
    )

    assert_array_almost_equal(FA.max(), expected_fa, decimal=2)


@set_random_number_generator(1980)
def test_add_noise(rng):
    N = 50
    S0 = 100

    options = {
        "func": f,
        "t": np.linspace(0, 2 * np.pi, N),
        "datashape": (10, 10, 10, len(bvals)),
        "origin": (5, 5, 5),
        "scale": (3, 3, 3),
        "angles": np.linspace(0, 2 * np.pi, 16),
        "radii": np.linspace(0.2, 2, 6),
        "S0": S0,
        "rng": rng,
    }

    vol = orbital_phantom(gtab=gtab, **options)

    for snr in [10, 20, 30, 50]:
        vol_noise = orbital_phantom(gtab=gtab, snr=snr, **options)

        sigma = S0 / snr

        assert_(np.abs(np.var(vol_noise - vol) - sigma**2) < 1)



from os.path import join as pjoin
from tempfile import TemporaryDirectory
import warnings

import numpy as np
import numpy.testing as npt
import pytest

from dipy.core.gradients import generate_bvecs
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti_data
from dipy.reconst import mapmri
from dipy.workflows.reconst import ReconstMAPMRIFlow


def test_reconst_mmri_laplacian():
    reconst_mmri_core(ReconstMAPMRIFlow, lap=True, pos=False)


def test_reconst_mmri_none():
    reconst_mmri_core(ReconstMAPMRIFlow, lap=False, pos=False)


@pytest.mark.skipif(not mapmri.have_cvxpy, reason="Requires CVXPY")
def test_reconst_mmri_both():
    reconst_mmri_core(ReconstMAPMRIFlow, lap=True, pos=True)


@pytest.mark.skipif(not mapmri.have_cvxpy, reason="Requires CVXPY")
def test_reconst_mmri_positivity():
    reconst_mmri_core(ReconstMAPMRIFlow, lap=True, pos=False)


def reconst_mmri_core(flow, lap, pos):
    with TemporaryDirectory() as out_dir:
        data_path, bval_path, bvec_path = get_fnames(name="small_25")
        volume = load_nifti_data(data_path)

        mmri_flow = flow()

        msg = "Optimization did not find a solution"
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message=msg, category=UserWarning)
            mmri_flow.run(
                data_files=data_path,
                bvals_files=bval_path,
                bvecs_files=bvec_path,
                small_delta=0.0129,
                big_delta=0.0218,
                laplacian=lap,
                positivity=pos,
                out_dir=out_dir,
            )

        for out_name in [
            "out_rtop",
            "out_lapnorm",
            "out_msd",
            "out_qiv",
            "out_rtap",
            "out_rtpp",
            "out_ng",
            "out_parng",
            "out_perng",
        ]:
            out_path = mmri_flow.last_generated_outputs[out_name]
            out_data = load_nifti_data(out_path)
            npt.assert_equal(out_data.shape, volume.shape[:-1])

        bvals, bvecs = read_bvals_bvecs(bval_path, bvec_path)
        bvals[0] = 5.0
        bvecs = generate_bvecs(len(bvals))
        tmp_bval_path = pjoin(out_dir, "tmp.bval")
        tmp_bvec_path = pjoin(out_dir, "tmp.bvec")
        np.savetxt(tmp_bval_path, bvals)
        np.savetxt(tmp_bvec_path, bvecs.T)
        mmri_flow._force_overwrite = True
        with npt.assert_raises(BaseException):
            npt.assert_warns(
                UserWarning,
                mmri_flow.run,
                data_path,
                tmp_bval_path,
                tmp_bvec_path,
                small_delta=0.0129,
                big_delta=0.0218,
                laplacian=lap,
                positivity=pos,
                out_dir=out_dir,
            )



from io import BytesIO
import pickle
from random import randint
import warnings

import numpy as np
from numpy.testing import (
    assert_,
    assert_almost_equal,
    assert_array_almost_equal,
    assert_array_equal,
    assert_equal,
    assert_raises,
    assert_warns,
)

from dipy.core.gradients import GradientTable, gradient_table
from dipy.core.sphere import HemiSphere, unit_icosahedron
from dipy.core.sphere_stats import angular_similarity
from dipy.core.subdivide_octahedron import create_unit_hemisphere
from dipy.data import default_sphere, get_fnames, get_sphere
from dipy.direction.peaks import (
    peak_directions,
    peak_directions_nl,
    peaks_from_model,
    peaks_from_positions,
    reshape_peaks_for_visualization,
)
from dipy.direction.pmf import SHCoeffPmfGen, SimplePmfGen
from dipy.io.gradients import read_bvals_bvecs
from dipy.reconst.odf import OdfFit, OdfModel, gfa
from dipy.reconst.shm import CsaOdfModel, descoteaux07_legacy_msg, tournier07_legacy_msg
from dipy.sims.voxel import multi_tensor, multi_tensor_odf
from dipy.testing.decorators import set_random_number_generator
from dipy.tracking.utils import seeds_from_mask


def test_peak_directions_nl():
    def discrete_eval(sphere):
        return abs(sphere.vertices).sum(-1)

    directions, values = peak_directions_nl(discrete_eval)
    assert_equal(directions.shape, (4, 3))
    assert_array_almost_equal(abs(directions), 1 / np.sqrt(3))
    assert_array_equal(values, abs(directions).sum(-1))

    # Test using a different sphere
    sphere = unit_icosahedron.subdivide(n=4)
    directions, values = peak_directions_nl(discrete_eval, sphere=sphere)
    assert_equal(directions.shape, (4, 3))
    assert_array_almost_equal(abs(directions), 1 / np.sqrt(3))
    assert_array_equal(values, abs(directions).sum(-1))

    # Test the relative_peak_threshold
    def discrete_eval(sphere):
        A = abs(sphere.vertices).sum(-1)
        x, y, z = sphere.vertices.T
        B = 1 + (x * z > 0) + 2 * (y * z > 0)
        return A * B

    directions, values = peak_directions_nl(discrete_eval, relative_peak_threshold=0.01)
    assert_equal(directions.shape, (4, 3))

    directions, values = peak_directions_nl(discrete_eval, relative_peak_threshold=0.3)
    assert_equal(directions.shape, (3, 3))

    directions, values = peak_directions_nl(discrete_eval, relative_peak_threshold=0.6)
    assert_equal(directions.shape, (2, 3))

    directions, values = peak_directions_nl(discrete_eval, relative_peak_threshold=0.8)
    assert_equal(directions.shape, (1, 3))
    assert_almost_equal(values, 4 * 3 / np.sqrt(3))

    # Test odfs with large areas of zero
    def discrete_eval(sphere):
        A = abs(sphere.vertices).sum(-1)
        x, y, z = sphere.vertices.T
        B = (x * z > 0) + 2 * (y * z > 0)
        return A * B

    directions, values = peak_directions_nl(discrete_eval, relative_peak_threshold=0.0)
    assert_equal(directions.shape, (3, 3))

    directions, values = peak_directions_nl(discrete_eval, relative_peak_threshold=0.6)
    assert_equal(directions.shape, (2, 3))

    directions, values = peak_directions_nl(discrete_eval, relative_peak_threshold=0.8)
    assert_equal(directions.shape, (1, 3))
    assert_almost_equal(values, 3 * 3 / np.sqrt(3))


_sphere = create_unit_hemisphere(recursion_level=4)
_odf = (_sphere.vertices * [1, 2, 3]).sum(-1)
_gtab = GradientTable(np.ones((64, 3)))


class SimpleOdfModel(OdfModel):
    sphere = _sphere

    def fit(self, data):
        fit = SimpleOdfFit(self, data)
        fit.model = self
        return fit


class SimpleOdfFit(OdfFit):
    def odf(self, sphere=None):
        if sphere is None:
            sphere = self.model.sphere

        # Use ascontiguousarray to work around a bug in NumPy
        return np.ascontiguousarray((sphere.vertices * [1, 2, 3]).sum(-1))


def test_OdfFit():
    m = SimpleOdfModel(_gtab)
    f = m.fit(None)
    odf = f.odf(_sphere)
    assert_equal(len(odf), len(_sphere.theta))


def test_peak_directions():
    model = SimpleOdfModel(_gtab)
    fit = model.fit(None)
    odf = fit.odf()

    argmax = odf.argmax()
    mx = odf.max()
    sphere = fit.model.sphere

    # Only one peak
    direction, val, ind = peak_directions(
        odf, sphere, relative_peak_threshold=0.5, min_separation_angle=45
    )
    dir_e = sphere.vertices[[argmax]]
    assert_array_equal(ind, [argmax])
    assert_array_equal(val, odf[ind])
    assert_array_equal(direction, dir_e)

    odf[0] = mx * 0.9
    # Two peaks, relative_threshold
    direction, val, ind = peak_directions(
        odf, sphere, relative_peak_threshold=1.0, min_separation_angle=0
    )
    dir_e = sphere.vertices[[argmax]]
    assert_array_equal(direction, dir_e)
    assert_array_equal(ind, [argmax])
    assert_array_equal(val, odf[ind])
    direction, val, ind = peak_directions(
        odf, sphere, relative_peak_threshold=0.8, min_separation_angle=0
    )
    dir_e = sphere.vertices[[argmax, 0]]
    assert_array_equal(direction, dir_e)
    assert_array_equal(ind, [argmax, 0])
    assert_array_equal(val, odf[ind])

    # Two peaks, angle_sep
    direction, val, ind = peak_directions(
        odf, sphere, relative_peak_threshold=0.0, min_separation_angle=90
    )
    dir_e = sphere.vertices[[argmax]]
    assert_array_equal(direction, dir_e)
    assert_array_equal(ind, [argmax])
    assert_array_equal(val, odf[ind])
    direction, val, ind = peak_directions(
        odf, sphere, relative_peak_threshold=0.0, min_separation_angle=0
    )
    dir_e = sphere.vertices[[argmax, 0]]
    assert_array_equal(direction, dir_e)
    assert_array_equal(ind, [argmax, 0])
    assert_array_equal(val, odf[ind])


def _create_mt_sim(mevals, angles, fractions, S0, SNR, half_sphere=False):
    _, fbvals, fbvecs = get_fnames(name="small_64D")

    bvals, bvecs = read_bvals_bvecs(fbvals, fbvecs)

    gtab = gradient_table(bvals, bvecs=bvecs)

    S, sticks = multi_tensor(
        gtab, mevals, S0=S0, angles=angles, fractions=fractions, snr=SNR
    )

    sphere = get_sphere(name="symmetric724").subdivide(n=2)

    if half_sphere:
        sphere = HemiSphere.from_sphere(sphere)

    odf_gt = multi_tensor_odf(
        sphere.vertices, mevals, angles=angles, fractions=fractions
    )

    return odf_gt, sticks, sphere


def test_peak_directions_thorough():
    # two equal fibers (creating a very sharp odf)
    mevals = np.array([[0.0025, 0.0003, 0.0003], [0.0025, 0.0003, 0.0003]])
    angles = [(0, 0), (45, 0)]
    fractions = [50, 50]
    odf_gt, sticks, sphere = _create_mt_sim(mevals, angles, fractions, 100, None)

    directions, values, indices = peak_directions(
        odf_gt, sphere, relative_peak_threshold=0.5, min_separation_angle=25.0
    )
    assert_almost_equal(angular_similarity(directions, sticks), 2, 2)

    # two unequal fibers
    fractions = [75, 25]
    odf_gt, sticks, sphere = _create_mt_sim(mevals, angles, fractions, 100, None)

    directions, values, indices = peak_directions(
        odf_gt, sphere, relative_peak_threshold=0.5, min_separation_angle=25.0
    )
    assert_almost_equal(angular_similarity(directions, sticks), 1, 2)

    directions, values, indices = peak_directions(
        odf_gt, sphere, relative_peak_threshold=0.20, min_separation_angle=25.0
    )
    assert_almost_equal(angular_similarity(directions, sticks), 2, 2)

    # two equal fibers short angle (simulating very sharp ODF)
    mevals = np.array(([0.0045, 0.0003, 0.0003], [0.0045, 0.0003, 0.0003]))
    fractions = [50, 50]
    angles = [(0, 0), (20, 0)]
    odf_gt, sticks, sphere = _create_mt_sim(mevals, angles, fractions, 100, None)

    directions, values, indices = peak_directions(
        odf_gt, sphere, relative_peak_threshold=0.5, min_separation_angle=25.0
    )
    assert_almost_equal(angular_similarity(directions, sticks), 1, 2)

    directions, values, indices = peak_directions(
        odf_gt, sphere, relative_peak_threshold=0.5, min_separation_angle=15.0
    )

    assert_almost_equal(angular_similarity(directions, sticks), 2, 2)

    # 1 fiber
    mevals = np.array([[0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003]])
    fractions = [50, 50]
    angles = [(15, 0), (15, 0)]
    odf_gt, sticks, sphere = _create_mt_sim(mevals, angles, fractions, 100, None)

    directions, values, indices = peak_directions(
        odf_gt, sphere, relative_peak_threshold=0.5, min_separation_angle=15.0
    )
    assert_almost_equal(angular_similarity(directions, sticks), 1, 2)

    AE = np.rad2deg(np.arccos(np.dot(directions[0], sticks[0])))
    assert_(abs(AE) < 2.0 or abs(AE - 180) < 2.0)

    # two equal fibers and one small noisy one
    mevals = np.array(
        [[0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003]]
    )
    angles = [(0, 0), (45, 0), (90, 0)]
    fractions = [45, 45, 10]
    odf_gt, sticks, sphere = _create_mt_sim(mevals, angles, fractions, 100, None)

    directions, values, indices = peak_directions(
        odf_gt, sphere, relative_peak_threshold=0.5, min_separation_angle=25.0
    )
    assert_almost_equal(angular_similarity(directions, sticks), 2, 2)

    # two equal fibers and one faulty
    mevals = np.array(
        [[0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003]]
    )
    angles = [(0, 0), (45, 0), (60, 0)]
    fractions = [45, 45, 10]
    odf_gt, sticks, sphere = _create_mt_sim(mevals, angles, fractions, 100, None)

    directions, values, indices = peak_directions(
        odf_gt, sphere, relative_peak_threshold=0.5, min_separation_angle=25.0
    )
    assert_almost_equal(angular_similarity(directions, sticks), 2, 2)

    # two equal fibers and one very very annoying one
    mevals = np.array(
        [[0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003]]
    )
    angles = [(0, 0), (45, 0), (60, 0)]
    fractions = [40, 40, 20]
    odf_gt, sticks, sphere = _create_mt_sim(mevals, angles, fractions, 100, None)

    directions, values, indices = peak_directions(
        odf_gt, sphere, relative_peak_threshold=0.5, min_separation_angle=25.0
    )
    assert_almost_equal(angular_similarity(directions, sticks), 2, 2)

    # three peaks and one faulty
    mevals = np.array(
        [
            [0.0015, 0.0003, 0.0003],
            [0.0015, 0.0003, 0.0003],
            [0.0015, 0.0003, 0.0003],
            [0.0015, 0.0003, 0.0003],
        ]
    )
    angles = [(0, 0), (45, 0), (90, 0), (90, 45)]
    fractions = [35, 35, 20, 10]
    odf_gt, sticks, sphere = _create_mt_sim(mevals, angles, fractions, 100, None)

    directions, values, indices = peak_directions(
        odf_gt, sphere, relative_peak_threshold=0.5, min_separation_angle=25.0
    )
    assert_almost_equal(angular_similarity(directions, sticks), 3, 2)

    # four peaks
    mevals = np.array(
        [
            [0.0015, 0.0003, 0.0003],
            [0.0015, 0.0003, 0.0003],
            [0.0015, 0.0003, 0.0003],
            [0.0015, 0.0003, 0.0003],
        ]
    )
    angles = [(0, 0), (45, 0), (90, 0), (90, 45)]
    fractions = [25, 25, 25, 25]
    odf_gt, sticks, sphere = _create_mt_sim(mevals, angles, fractions, 100, None)

    directions, values, indices = peak_directions(
        odf_gt, sphere, relative_peak_threshold=0.15, min_separation_angle=5.0
    )
    assert_almost_equal(angular_similarity(directions, sticks), 4, 2)

    # four difficult peaks
    mevals = np.array(
        [
            [0.0015, 0.0003, 0.0003],
            [0.0015, 0.0003, 0.0003],
            [0.0015, 0.0003, 0.0003],
            [0.0015, 0.0003, 0.0003],
        ]
    )
    angles = [(0, 0), (45, 0), (90, 0), (90, 45)]
    fractions = [30, 30, 20, 20]
    odf_gt, sticks, sphere = _create_mt_sim(mevals, angles, fractions, 100, None)

    directions, values, indices = peak_directions(
        odf_gt, sphere, relative_peak_threshold=0, min_separation_angle=0
    )
    assert_almost_equal(angular_similarity(directions, sticks), 4, 1)

    # test the asymmetric case
    directions, values, indices = peak_directions(
        odf_gt,
        sphere,
        relative_peak_threshold=0,
        min_separation_angle=0,
        is_symmetric=False,
    )
    expected = np.concatenate([sticks, -sticks], axis=0)
    assert_almost_equal(angular_similarity(directions, expected), 8, 1)

    odf_gt, sticks, hsphere = _create_mt_sim(
        mevals, angles, fractions, 100, None, half_sphere=True
    )

    directions, values, indices = peak_directions(
        odf_gt, hsphere, relative_peak_threshold=0, min_separation_angle=0
    )
    assert_equal(angular_similarity(directions, sticks) < 4, True)

    # four peaks and one them quite small
    fractions = [35, 35, 20, 10]

    odf_gt, sticks, sphere = _create_mt_sim(mevals, angles, fractions, 100, None)

    directions, values, indices = peak_directions(
        odf_gt, sphere, relative_peak_threshold=0, min_separation_angle=0
    )
    assert_equal(angular_similarity(directions, sticks) < 4, True)

    odf_gt, sticks, hsphere = _create_mt_sim(
        mevals, angles, fractions, 100, None, half_sphere=True
    )

    directions, values, indices = peak_directions(
        odf_gt, hsphere, relative_peak_threshold=0, min_separation_angle=0
    )
    assert_equal(angular_similarity(directions, sticks) < 4, True)

    # isotropic case
    mevals = np.array([[0.0015, 0.0015, 0.0015]])
    angles = [(0, 0)]
    fractions = [100.0]
    odf_gt, sticks, sphere = _create_mt_sim(mevals, angles, fractions, 100, None)

    directions, values, indices = peak_directions(
        odf_gt, sphere, relative_peak_threshold=0.5, min_separation_angle=25.0
    )
    assert_equal(len(values) > 10, True)


def test_difference_with_minmax():
    # Show difference with and without minmax normalization
    # we create an odf here with 3 main peaks, 1 small sharp unwanted peak
    # (noise) and an isotropic compartment.
    mevals = np.array(
        [
            [0.0015, 0.0003, 0.0003],
            [0.0015, 0.0003, 0.0003],
            [0.0015, 0.0003, 0.0003],
            [0.0015, 0.00005, 0.00005],
            [0.0015, 0.0015, 0.0015],
        ]
    )
    angles = [(0, 0), (45, 0), (90, 0), (90, 90), (0, 0)]
    fractions = [20, 20, 10, 1, 100 - 20 - 20 - 10 - 1]
    odf_gt, sticks, sphere = _create_mt_sim(mevals, angles, fractions, 100, None)

    # We will show that when the minmax normalization is used we can remove
    # the noisy peak using a lower threshold.

    odf_gt_minmax = (odf_gt - odf_gt.min()) / (odf_gt.max() - odf_gt.min())

    _, values_1, _ = peak_directions(
        odf_gt, sphere, relative_peak_threshold=0.30, min_separation_angle=25.0
    )

    assert_equal(len(values_1), 3)

    _, values_2, _ = peak_directions(
        odf_gt_minmax, sphere, relative_peak_threshold=0.30, min_separation_angle=25.0
    )

    assert_equal(len(values_2), 3)

    # Setting the smallest value of the odf to zero is like running
    # peak_directions without the odf_min correction.
    odf_gt[odf_gt.argmin()] = 0.0
    _, values_3, _ = peak_directions(
        odf_gt,
        sphere,
        relative_peak_threshold=0.30,
        min_separation_angle=25.0,
    )

    assert_equal(len(values_3), 4)

    # we show here that to actually get that noisy peak out we need to
    # increase the peak threshold considerably
    directions, values_4, indices = peak_directions(
        odf_gt,
        sphere,
        relative_peak_threshold=0.60,
        min_separation_angle=25.0,
    )

    assert_equal(len(values_4), 3)
    assert_almost_equal(values_1, values_4)


@set_random_number_generator()
def test_degenerate_cases(rng):
    sphere = default_sphere

    # completely isotropic and degenerate case
    odf = np.zeros(sphere.vertices.shape[0])
    directions, values, indices = peak_directions(
        odf, sphere, relative_peak_threshold=0.5, min_separation_angle=25
    )
    print(directions, values, indices)

    assert_equal(len(values), 0)
    assert_equal(len(directions), 0)
    assert_equal(len(indices), 0)

    odf = np.zeros(sphere.vertices.shape[0])
    odf[0] = 0.020
    odf[1] = 0.018

    directions, values, indices = peak_directions(
        odf, sphere, relative_peak_threshold=0.5, min_separation_angle=25
    )
    print(directions, values, indices)

    assert_equal(values[0], 0.02)

    odf = -np.ones(sphere.vertices.shape[0])
    directions, values, indices = peak_directions(
        odf, sphere, relative_peak_threshold=0.5, min_separation_angle=25
    )
    print(directions, values, indices)

    assert_equal(len(values), 0)

    odf = np.zeros(sphere.vertices.shape[0])
    odf[0] = 0.020
    odf[1] = 0.018
    odf[2] = -0.018

    directions, values, indices = peak_directions(
        odf, sphere, relative_peak_threshold=0.5, min_separation_angle=25
    )
    assert_equal(values[0], 0.02)

    odf = np.ones(sphere.vertices.shape[0])
    odf += 0.1 * rng.random(odf.shape[0])
    directions, values, indices = peak_directions(
        odf, sphere, relative_peak_threshold=0.5, min_separation_angle=25
    )
    assert_(all(values > values[0] * 0.5))
    assert_array_equal(values, odf[indices])

    odf = np.ones(sphere.vertices.shape[0])
    odf[1:] = np.finfo(float).eps * rng.random(odf.shape[0] - 1)
    directions, values, indices = peak_directions(
        odf, sphere, relative_peak_threshold=0.5, min_separation_angle=25
    )

    assert_equal(values[0], 1)
    assert_equal(len(values), 1)


def test_peaksFromModel():
    data = np.zeros((10, 2))

    for sphere in [_sphere, get_sphere(name="symmetric642")]:
        # Test basic case
        model = SimpleOdfModel(_gtab)
        _odf = (sphere.vertices * [1, 2, 3]).sum(-1)
        odf_argmax = _odf.argmax()
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )
            pam = peaks_from_model(model, data, sphere, 0.5, 45, normalize_peaks=True)

        assert_array_equal(pam.gfa, gfa(_odf))
        assert_array_equal(pam.peak_values[:, 0], 1.0)
        assert_array_equal(pam.peak_values[:, 1:], 0.0)
        mn, mx = _odf.min(), _odf.max()
        assert_array_equal(pam.qa[:, 0], (mx - mn) / mx)
        assert_array_equal(pam.qa[:, 1:], 0.0)
        assert_array_equal(pam.peak_indices[:, 0], odf_argmax)
        assert_array_equal(pam.peak_indices[:, 1:], -1)

        # Test that odf array matches and is right shape
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )
            pam = peaks_from_model(model, data, sphere, 0.5, 45, return_odf=True)
        expected_shape = (len(data), len(_odf))
        assert_equal(pam.odf.shape, expected_shape)
        assert_((_odf == pam.odf).all())
        assert_array_equal(pam.peak_values[:, 0], _odf.max())

        # Test mask
        mask = (np.arange(10) % 2) == 1

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )
            pam = peaks_from_model(
                model, data, sphere, 0.5, 45, mask=mask, normalize_peaks=True
            )
        assert_array_equal(pam.gfa[~mask], 0)
        assert_array_equal(pam.qa[~mask], 0)
        assert_array_equal(pam.peak_values[~mask], 0)
        assert_array_equal(pam.peak_indices[~mask], -1)

        assert_array_equal(pam.gfa[mask], gfa(_odf))
        assert_array_equal(pam.peak_values[mask, 0], 1.0)
        assert_array_equal(pam.peak_values[mask, 1:], 0.0)
        mn, mx = _odf.min(), _odf.max()
        assert_array_equal(pam.qa[mask, 0], (mx - mn) / mx)
        assert_array_equal(pam.qa[mask, 1:], 0.0)
        assert_array_equal(pam.peak_indices[mask, 0], odf_argmax)
        assert_array_equal(pam.peak_indices[mask, 1:], -1)

        # Test serialization and deserialization:
        for normalize_peaks in [True, False]:
            for return_odf in [True, False]:
                for return_sh in [True, False]:
                    with warnings.catch_warnings():
                        warnings.filterwarnings(
                            "ignore",
                            message=descoteaux07_legacy_msg,
                            category=PendingDeprecationWarning,
                        )
                        pam = peaks_from_model(
                            model,
                            data,
                            sphere,
                            0.5,
                            45,
                            normalize_peaks=normalize_peaks,
                            return_odf=return_odf,
                            return_sh=return_sh,
                        )

                    b = BytesIO()
                    pickle.dump(pam, b)
                    b.seek(0)
                    new_pam = pickle.load(b)
                    b.close()

                    for attr in [
                        "peak_dirs",
                        "peak_values",
                        "peak_indices",
                        "gfa",
                        "qa",
                        "shm_coeff",
                        "B",
                        "odf",
                    ]:
                        assert_array_equal(getattr(pam, attr), getattr(new_pam, attr))
                        assert_array_equal(pam.sphere.vertices, new_pam.sphere.vertices)


def test_peaksFromModelParallel():
    SNR = 100
    S0 = 100

    _, fbvals, fbvecs = get_fnames(name="small_64D")

    bvals, bvecs = read_bvals_bvecs(fbvals, fbvecs)

    gtab = gradient_table(bvals, bvecs=bvecs)
    mevals = np.array(([0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003]))

    data, _ = multi_tensor(
        gtab, mevals, S0=S0, angles=[(0, 0), (60, 0)], fractions=[50, 50], snr=SNR
    )

    for sphere in [_sphere, default_sphere]:
        # test equality with/without multiprocessing
        model = SimpleOdfModel(gtab)
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )
            pam_multi = peaks_from_model(
                model,
                data,
                sphere,
                relative_peak_threshold=0.5,
                min_separation_angle=45,
                normalize_peaks=True,
                return_odf=True,
                return_sh=True,
                parallel=True,
            )

            pam_single = peaks_from_model(
                model,
                data,
                sphere,
                relative_peak_threshold=0.5,
                min_separation_angle=45,
                normalize_peaks=True,
                return_odf=True,
                return_sh=True,
                parallel=False,
            )

            pam_multi_inv1 = peaks_from_model(
                model,
                data,
                sphere,
                relative_peak_threshold=0.5,
                min_separation_angle=45,
                normalize_peaks=True,
                return_odf=True,
                return_sh=True,
                parallel=True,
                num_processes=-1,
            )

            pam_multi_inv2 = peaks_from_model(
                model,
                data,
                sphere,
                relative_peak_threshold=0.5,
                min_separation_angle=45,
                normalize_peaks=True,
                return_odf=True,
                return_sh=True,
                parallel=True,
                num_processes=-2,
            )

        for pam in [pam_multi, pam_multi_inv1, pam_multi_inv2]:
            assert_equal(pam.gfa.dtype, pam_single.gfa.dtype)
            assert_equal(pam.gfa.shape, pam_single.gfa.shape)
            assert_array_almost_equal(pam.gfa, pam_single.gfa)

            assert_equal(pam.qa.dtype, pam_single.qa.dtype)
            assert_equal(pam.qa.shape, pam_single.qa.shape)
            assert_array_almost_equal(pam.qa, pam_single.qa)

            assert_equal(pam.peak_values.dtype, pam_single.peak_values.dtype)
            assert_equal(pam.peak_values.shape, pam_single.peak_values.shape)
            assert_array_almost_equal(pam.peak_values, pam_single.peak_values)

            assert_equal(pam.peak_indices.dtype, pam_single.peak_indices.dtype)
            assert_equal(pam.peak_indices.shape, pam_single.peak_indices.shape)
            assert_array_equal(pam.peak_indices, pam_single.peak_indices)

            assert_equal(pam.peak_dirs.dtype, pam_single.peak_dirs.dtype)
            assert_equal(pam.peak_dirs.shape, pam_single.peak_dirs.shape)
            assert_array_almost_equal(pam.peak_dirs, pam_single.peak_dirs)

            assert_equal(pam.shm_coeff.dtype, pam_single.shm_coeff.dtype)
            assert_equal(pam.shm_coeff.shape, pam_single.shm_coeff.shape)
            assert_array_almost_equal(pam.shm_coeff, pam_single.shm_coeff)

            assert_equal(pam.odf.dtype, pam_single.odf.dtype)
            assert_equal(pam.odf.shape, pam_single.odf.shape)
            assert_array_almost_equal(pam.odf, pam_single.odf)


def test_peaks_shm_coeff():
    SNR = 100
    S0 = 100

    _, fbvals, fbvecs = get_fnames(name="small_64D")

    sphere = default_sphere

    bvals, bvecs = read_bvals_bvecs(fbvals, fbvecs)

    gtab = gradient_table(bvals, bvecs=bvecs)
    mevals = np.array(([0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003]))

    data, _ = multi_tensor(
        gtab, mevals, S0=S0, angles=[(0, 0), (60, 0)], fractions=[50, 50], snr=SNR
    )

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        model = CsaOdfModel(gtab, 4)
        pam = peaks_from_model(
            model, data[None, :], sphere, 0.5, 45, return_odf=True, return_sh=True
        )
    # Test that spherical harmonic coefficients return back correctly
    odf2 = np.dot(pam.shm_coeff, pam.B)
    assert_array_almost_equal(pam.odf, odf2)
    assert_equal(pam.shm_coeff.shape[-1], 45)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        pam = peaks_from_model(
            model, data[None, :], sphere, 0.5, 45, return_odf=True, return_sh=False
        )
    assert_equal(pam.shm_coeff, None)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message=tournier07_legacy_msg, category=PendingDeprecationWarning
        )
        pam = peaks_from_model(
            model,
            data[None, :],
            sphere,
            0.5,
            45,
            return_odf=True,
            return_sh=True,
            sh_basis_type="tournier07",
        )

    odf2 = np.dot(pam.shm_coeff, pam.B)
    assert_array_almost_equal(pam.odf, odf2)


@set_random_number_generator()
def test_reshape_peaks_for_visualization(rng):
    data1 = rng.standard_normal((10, 5, 3)).astype("float32")
    data2 = rng.standard_normal((10, 2, 5, 3)).astype("float32")
    data3 = rng.standard_normal((10, 2, 12, 5, 3)).astype("float32")

    data1_reshape = reshape_peaks_for_visualization(data1)
    data2_reshape = reshape_peaks_for_visualization(data2)
    data3_reshape = reshape_peaks_for_visualization(data3)

    assert_array_equal(data1_reshape.shape, (10, 15))
    assert_array_equal(data2_reshape.shape, (10, 2, 15))
    assert_array_equal(data3_reshape.shape, (10, 2, 12, 15))

    assert_array_equal(data1_reshape.reshape(10, 5, 3), data1)
    assert_array_equal(data2_reshape.reshape(10, 2, 5, 3), data2)
    assert_array_equal(data3_reshape.reshape(10, 2, 12, 5, 3), data3)


def test_peaks_from_positions():
    thresh = 0.5
    min_angle = 25
    npeaks = 5

    _, fbvals, fbvecs = get_fnames(name="small_64D")
    bvals, bvecs = read_bvals_bvecs(fbvals, fbvecs)
    gtab = gradient_table(bvals, bvecs=bvecs)
    mevals = np.array(([0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003]))
    voxels = []
    for _ in range(27):
        v, _ = multi_tensor(
            gtab,
            mevals,
            S0=100,
            angles=[(0, 0), (randint(0, 90), randint(0, 90))],
            fractions=[50, 50],
            snr=10,
        )
        voxels.append(v)
    data = np.array(voxels).reshape((3, 3, 3, -1))

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        model = CsaOdfModel(gtab, 8)
        pam = peaks_from_model(
            model,
            data,
            default_sphere,
            return_odf=True,
            return_sh=True,
            legacy=True,
            npeaks=npeaks,
            relative_peak_threshold=thresh,
            min_separation_angle=min_angle,
        )

    mask = np.ones((3, 3, 3))
    affine = np.eye(4)
    positions = seeds_from_mask(mask, affine)

    # test the peaks at each voxel using int coordinates
    peaks = peaks_from_positions(
        positions,
        pam.odf,
        default_sphere,
        affine,
        relative_peak_threshold=thresh,
        min_separation_angle=min_angle,
        npeaks=npeaks,
    )
    peaks = np.array(peaks).reshape((3, 3, 3, 5, 3))
    assert_array_almost_equal(pam.peak_dirs, peaks)

    # test the peaks at each voxel using float coordinates
    peaks = peaks_from_positions(
        positions.astype(float),
        pam.odf,
        default_sphere,
        affine,
        relative_peak_threshold=thresh,
        min_separation_angle=min_angle,
        npeaks=npeaks,
    )
    peaks = np.array(peaks).reshape((3, 3, 3, 5, 3))
    assert_array_almost_equal(pam.peak_dirs, peaks)

    # test the peaks at each voxel using double coordinates
    peaks = peaks_from_positions(
        positions.astype(np.float64),
        pam.odf,
        default_sphere,
        affine,
        relative_peak_threshold=thresh,
        min_separation_angle=min_angle,
        npeaks=npeaks,
    )
    peaks = np.array(peaks).reshape((3, 3, 3, 5, 3))
    assert_array_almost_equal(pam.peak_dirs, peaks)

    # test the peaks at each voxel using SimplePmfGen
    pmf_gen = SimplePmfGen(pam.odf, default_sphere)
    peaks = peaks_from_positions(
        positions,
        odfs=None,
        sphere=None,
        affine=affine,
        pmf_gen=pmf_gen,
        relative_peak_threshold=thresh,
        min_separation_angle=min_angle,
        npeaks=npeaks,
    )
    peaks = np.array(peaks).reshape((3, 3, 3, 5, 3))
    assert_array_almost_equal(pam.peak_dirs, peaks, decimal=3)

    # test the peaks at each voxel using SHCoeffPmfGen
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        pmf_gen = SHCoeffPmfGen(
            pam.shm_coeff, default_sphere, basis_type="descoteaux07"
        )
    peaks = peaks_from_positions(
        positions,
        odfs=None,
        sphere=None,
        affine=affine,
        pmf_gen=pmf_gen,
        relative_peak_threshold=thresh,
        min_separation_angle=min_angle,
        npeaks=npeaks,
    )
    peaks = np.array(peaks).reshape((3, 3, 3, 5, 3))
    assert_array_almost_equal(pam.peak_dirs, peaks, decimal=3)

    # test the peaks with a full sphere
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        pam_full_sphere = peaks_from_model(
            model,
            data,
            get_sphere(name="symmetric362"),
            return_odf=True,
            return_sh=True,
            legacy=True,
            npeaks=npeaks,
            relative_peak_threshold=thresh,
            min_separation_angle=min_angle,
        )
    pmf_gen = SimplePmfGen(pam_full_sphere.odf, get_sphere(name="symmetric362"))
    peaks = peaks_from_positions(
        positions,
        odfs=None,
        sphere=None,
        affine=affine,
        pmf_gen=pmf_gen,
        relative_peak_threshold=thresh,
        min_separation_angle=min_angle,
        npeaks=npeaks,
    )
    peaks = np.array(peaks).reshape((3, 3, 3, 5, 3))
    assert_array_almost_equal(pam_full_sphere.peak_dirs, peaks, decimal=3)

    # test the peaks extraction at the mid point between 2 voxels
    odfs = [pam.odf[0, 0, 0], pam.odf[0, 0, 0]]
    odfs = np.array(odfs).reshape((2, 1, 1, -1))
    positions = np.array([[0.0, 0, 0], [0.5, 0, 0], [1.0, 0, 0]])
    peaks = peaks_from_positions(
        positions,
        odfs,
        default_sphere,
        affine,
        relative_peak_threshold=thresh,
        min_separation_angle=min_angle,
        npeaks=npeaks,
    )
    assert_array_equal(peaks[0], peaks[1])
    assert_array_equal(peaks[0], peaks[2])

    # test with none identity affine
    positions = seeds_from_mask(mask, affine)
    peaks_eye = peaks_from_positions(
        positions,
        pam.odf,
        default_sphere,
        affine,
        relative_peak_threshold=thresh,
        min_separation_angle=min_angle,
        npeaks=npeaks,
    )

    affine[:3, :3] = np.random.random((3, 3))
    positions = seeds_from_mask(mask, affine)

    peaks = peaks_from_positions(
        positions,
        pam.odf,
        default_sphere,
        affine,
        relative_peak_threshold=thresh,
        min_separation_angle=min_angle,
        npeaks=npeaks,
    )
    assert_array_almost_equal(peaks_eye, peaks)

    # test with invalid seed coordinates
    affine = np.eye(4)
    positions = np.array([[0, -1, 0], [0.1, -0.1, 0.1]])

    assert_raises(
        IndexError,
        peaks_from_positions,
        positions,
        pam.odf,
        default_sphere,
        affine,
    )

    affine = np.eye(4) * 10
    positions = np.array([[1, -1, 1]])
    positions = np.dot(positions, affine[:3, :3].T)
    positions += affine[:3, 3]
    assert_raises(
        IndexError,
        peaks_from_positions,
        positions,
        pam.odf,
        default_sphere,
        affine,
    )

    # test a warning is thrown when odfs and pmf_gen arguments are used
    assert_warns(
        UserWarning,
        peaks_from_positions,
        positions,
        pam.odf,
        default_sphere,
        affine,
        pmf_gen=pmf_gen,
        relative_peak_threshold=thresh,
        min_separation_angle=min_angle,
        npeaks=npeaks,
    )



"""Test scripts

Run scripts and check outputs
"""

"""

import glob
import os
import shutil
from tempfile import TemporaryDirectory

from os.path import (dirname, join as pjoin, abspath)

from dipy.testing import assert_true, assert_false
import numpy.testing as nt
import pytest

import nibabel as nib

from dipy.data import get_fnames

# Quickbundles command-line requires matplotlib:
try:
    import matplotlib
    no_mpl = False
except ImportError:
    no_mpl = True

from dipy.tests.scriptrunner import ScriptRunner

runner = ScriptRunner(
    script_sdir='bin',
    debug_print_var='NIPY_DEBUG_PRINT')
run_command = runner.run_command

DATA_PATH = abspath(pjoin(dirname(__file__), 'data'))


def test_dipy_peak_extraction():
    # test dipy_peak_extraction script
    cmd = 'dipy_peak_extraction'
    code, stdout, stderr = run_command(cmd, check_code=False)
    npt.assert_equal(code, 2)


def test_dipy_fit_tensor():
    # test dipy_fit_tensor script
    cmd = 'dipy_fit_tensor'
    code, stdout, stderr = run_command(cmd, check_code=False)
    npt.assert_equal(code, 2)


def test_dipy_sh_estimate():
    # test dipy_sh_estimate script
    cmd = 'dipy_sh_estimate'
    code, stdout, stderr = run_command(cmd, check_code=False)
    npt.assert_equal(code, 2)


def assert_image_shape_affine(filename, shape, affine):
    assert_true(os.path.isfile(filename))
    image = nib.load(filename)
    npt.assert_equal(image.shape, shape)
    nt.assert_array_almost_equal(image.affine, affine)


def test_dipy_fit_tensor_again():
    with TemporaryDirectory():
        dwi, bval, bvec = get_fnames(name="small_25")
        # Copy data to tmp directory
        shutil.copyfile(dwi, "small_25.nii.gz")
        shutil.copyfile(bval, "small_25.bval")
        shutil.copyfile(bvec, "small_25.bvec")
        # Call script
        cmd = ["dipy_fit_tensor", "--mask=none", "small_25.nii.gz"]
        out = run_command(cmd)
        npt.assert_equal(out[0], 0)
        # Get expected values
        img = nib.load("small_25.nii.gz")
        affine = img.affine
        shape = img.shape[:-1]
        # Check expected outputs
        assert_image_shape_affine("small_25_fa.nii.gz", shape, affine)
        assert_image_shape_affine("small_25_t2di.nii.gz", shape, affine)
        assert_image_shape_affine("small_25_dirFA.nii.gz", shape, affine)
        assert_image_shape_affine("small_25_ad.nii.gz", shape, affine)
        assert_image_shape_affine("small_25_md.nii.gz", shape, affine)
        assert_image_shape_affine("small_25_rd.nii.gz", shape, affine)

    with TemporaryDirectory():
        dwi, bval, bvec = get_fnames(name="small_25")
        # Copy data to tmp directory
        shutil.copyfile(dwi, "small_25.nii.gz")
        shutil.copyfile(bval, "small_25.bval")
        shutil.copyfile(bvec, "small_25.bvec")
        # Call script
        cmd = ["dipy_fit_tensor", "--save-tensor",
               "--mask=none", "small_25.nii.gz"]
        out = run_command(cmd)
        npt.assert_equal(out[0], 0)
        # Get expected values
        img = nib.load("small_25.nii.gz")
        affine = img.affine
        shape = img.shape[:-1]
        # Check expected outputs
        assert_image_shape_affine("small_25_fa.nii.gz", shape, affine)
        assert_image_shape_affine("small_25_t2di.nii.gz", shape, affine)
        assert_image_shape_affine("small_25_dirFA.nii.gz", shape, affine)
        assert_image_shape_affine("small_25_ad.nii.gz", shape, affine)
        assert_image_shape_affine("small_25_md.nii.gz", shape, affine)
        assert_image_shape_affine("small_25_rd.nii.gz", shape, affine)
        # small_25_tensor saves the tensor as a symmetric matrix following
        # the nifti standard.
        ten_shape = shape + (1, 6)
        assert_image_shape_affine("small_25_tensor.nii.gz", ten_shape,
                                  affine)


@pytest.mark.skipif(no_mpl)
def test_qb_commandline():
    with TemporaryDirectory():
        tracks_file = get_fnames(name='fornix')
        cmd = ["dipy_quickbundles", tracks_file, '--pkl_file', 'mypickle.pkl',
               '--out_file', 'tracks300.trk']
        out = run_command(cmd)
        npt.assert_equal(out[0], 0)

@pytest.mark.skipif(no_mpl)
def test_qb_commandline_output_path_handling():
    with TemporaryDirectory():
        # Create temporary subdirectory for input and for output
        os.mkdir('work')
        os.mkdir('output')

        os.chdir('work')
        tracks_file = get_fnames(name='fornix')

        # Need to specify an output directory with a "../" style path
        # to trigger old bug.
        cmd = ["dipy_quickbundles", tracks_file, '--pkl_file', 'mypickle.pkl',
               '--out_file', os.path.join('..', 'output', 'tracks300.trk')]
        out = run_command(cmd)
        npt.assert_equal(out[0], 0)

        # Make sure the files were created in the output directory
        os.chdir('../')
        output_files_list = glob.glob('output/tracks300_*.trk')
        assert_true(output_files_list)
"""



#!/usr/bin/python
"""Class and helper functions for fitting the EVAC+ model."""

import logging

import numpy as np

from dipy.align.reslice import reslice
from dipy.data import get_fnames
from dipy.nn.utils import (
    normalize,
    recover_img,
    set_logger_level,
    transform_img,
)
from dipy.segment.utils import remove_holes_and_islands
from dipy.testing.decorators import doctest_skip_parser, warning_for_keywords
from dipy.utils.deprecator import deprecated_params
from dipy.utils.optpkg import optional_package

tf, have_tf, _ = optional_package("tensorflow", min_version="2.18.0")
if have_tf:
    from tensorflow.keras.layers import (
        Add,
        Concatenate,
        Conv3D,
        Conv3DTranspose,
        Dropout,
        Layer,
        LayerNormalization,
        ReLU,
        Softmax,
    )
    from tensorflow.keras.models import Model
else:

    class Model:
        pass

    class Layer:
        pass

    logging.warning(
        "This model requires Tensorflow.\
                    Please install these packages using \
                    pip. If using mac, please refer to this \
                    link for installation. \
                    https://github.com/apple/tensorflow_macos"
    )

logging.basicConfig()
logger = logging.getLogger("EVAC+")


def prepare_img(image):
    """Function to prepare image for model input
    Specific to EVAC+

    Parameters
    ----------
    image : np.ndarray
        Input image

    Returns
    -------
    input_data : dict
    """
    input1 = np.moveaxis(image, -1, 0)
    input1 = np.expand_dims(input1, -1)

    input2, _ = reslice(image, np.eye(4), (1, 1, 1), (2, 2, 2))
    input2 = np.moveaxis(input2, -1, 0)
    input2 = np.expand_dims(input2, -1)

    input3, _ = reslice(image, np.eye(4), (1, 1, 1), (4, 4, 4))
    input3 = np.moveaxis(input3, -1, 0)
    input3 = np.expand_dims(input3, -1)

    input4, _ = reslice(image, np.eye(4), (1, 1, 1), (8, 8, 8))
    input4 = np.moveaxis(input4, -1, 0)
    input4 = np.expand_dims(input4, -1)

    input5, _ = reslice(image, np.eye(4), (1, 1, 1), (16, 16, 16))
    input5 = np.moveaxis(input5, -1, 0)
    input5 = np.expand_dims(input5, -1)

    input_data = {
        "input_1": input1,
        "input_2": input2,
        "input_3": input3,
        "input_4": input4,
        "input_5": input5,
    }

    return input_data


class Block(Layer):
    @warning_for_keywords()
    def __init__(
        self,
        out_channels,
        kernel_size,
        strides,
        padding,
        drop_r,
        n_layers,
        *,
        layer_type="down",
    ):
        super(Block, self).__init__()
        self.layer_list = []
        self.layer_list2 = []
        self.n_layers = n_layers
        for _ in range(n_layers):
            self.layer_list.append(
                Conv3D(out_channels, kernel_size, strides=strides, padding=padding)
            )
            self.layer_list.append(Dropout(drop_r))
            self.layer_list.append(LayerNormalization())
            self.layer_list.append(ReLU())
        if layer_type == "down":
            self.layer_list2.append(Conv3D(1, 2, strides=2, padding="same"))
            self.layer_list2.append(ReLU())
        elif layer_type == "up":
            self.layer_list2.append(Conv3DTranspose(1, 2, strides=2, padding="same"))
            self.layer_list2.append(ReLU())

        self.channel_sum = ChannelSum()
        self.add = Add()

    def call(self, input, passed):
        x = input
        for layer in self.layer_list:
            x = layer(x)

        x = self.channel_sum(x)
        fwd = self.add([x, passed])
        x = fwd

        for layer in self.layer_list2:
            x = layer(x)

        return fwd, x


class ChannelSum(Layer):
    def __init__(self):
        super(ChannelSum, self).__init__()

    def call(self, inputs):
        return tf.reduce_sum(inputs, axis=-1, keepdims=True)


@warning_for_keywords()
def init_model(*, model_scale=16):
    """Function to create model for EVAC+

    Parameters
    ----------
    model_scale : int, optional
        The scale of the model
        Should match the saved weights from fetcher
        Default is 16

    Returns
    -------
    model : tf.keras.Model
    """
    inputs = tf.keras.Input(shape=(128, 128, 128, 1), name="input_1")
    raw_input_2 = tf.keras.Input(shape=(64, 64, 64, 1), name="input_2")
    raw_input_3 = tf.keras.Input(shape=(32, 32, 32, 1), name="input_3")
    raw_input_4 = tf.keras.Input(shape=(16, 16, 16, 1), name="input_4")
    raw_input_5 = tf.keras.Input(shape=(8, 8, 8, 1), name="input_5")
    # Encode
    fwd1, x = Block(
        model_scale, kernel_size=5, strides=1, padding="same", drop_r=0.2, n_layers=1
    )(inputs, inputs)

    x = Concatenate()([x, raw_input_2])

    fwd2, x = Block(
        model_scale * 2,
        kernel_size=5,
        strides=1,
        padding="same",
        drop_r=0.5,
        n_layers=2,
    )(x, x)

    x = Concatenate()([x, raw_input_3])

    fwd3, x = Block(
        model_scale * 4,
        kernel_size=5,
        strides=1,
        padding="same",
        drop_r=0.5,
        n_layers=3,
    )(x, x)

    x = Concatenate()([x, raw_input_4])

    fwd4, x = Block(
        model_scale * 8,
        kernel_size=5,
        strides=1,
        padding="same",
        drop_r=0.5,
        n_layers=3,
    )(x, x)

    x = Concatenate()([x, raw_input_5])

    _, up = Block(
        model_scale * 16,
        kernel_size=5,
        strides=1,
        padding="same",
        drop_r=0.5,
        n_layers=3,
        layer_type="up",
    )(x, x)

    x = Concatenate()([fwd4, up])

    _, up = Block(
        model_scale * 8,
        kernel_size=5,
        strides=1,
        padding="same",
        drop_r=0.5,
        n_layers=3,
        layer_type="up",
    )(x, up)

    x = Concatenate()([fwd3, up])

    _, up = Block(
        model_scale * 4,
        kernel_size=5,
        strides=1,
        padding="same",
        drop_r=0.5,
        n_layers=3,
        layer_type="up",
    )(x, up)

    x = Concatenate()([fwd2, up])

    _, up = Block(
        model_scale * 2,
        kernel_size=5,
        strides=1,
        padding="same",
        drop_r=0.5,
        n_layers=2,
        layer_type="up",
    )(x, up)

    x = Concatenate()([fwd1, up])

    _, pred = Block(
        model_scale,
        kernel_size=5,
        strides=1,
        padding="same",
        drop_r=0.5,
        n_layers=1,
        layer_type="none",
    )(x, up)

    pred = Conv3D(2, 1, padding="same")(pred)
    output = Softmax(axis=-1)(pred)

    model = Model(
        {
            "input_1": inputs,
            "input_2": raw_input_2,
            "input_3": raw_input_3,
            "input_4": raw_input_4,
            "input_5": raw_input_5,
        },
        output[..., 0],
    )
    return model


class EVACPlus:
    """This class is intended for the EVAC+ model.

    The EVAC+ model :footcite:p:`Park2024` is a deep learning neural network for
    brain extraction. It uses a V-net architecture combined with
    multi-resolution input data, an additional conditional random field (CRF)
    recurrent layer and supplementary Dice loss term for this recurrent layer.

    References
    ----------
    .. footbibliography::
    """

    @doctest_skip_parser
    @warning_for_keywords()
    def __init__(self, *, verbose=False):
        """Model initialization

        The model was pre-trained for usage on
        brain extraction of T1 images.

        This model is designed to take as input
        a T1 weighted image.

        Parameters
        ----------
        verbose : bool, optional
            Whether to show information about the processing.
        """
        if not have_tf:
            raise tf()

        log_level = "INFO" if verbose else "CRITICAL"
        set_logger_level(log_level, logger)

        # EVAC+ network load

        self.model = init_model()
        self.fetch_default_weights()

    def fetch_default_weights(self):
        """Load the model pre-training weights to use for the fitting.
        While the user can load different weights, the function
        is mainly intended for the class function 'predict'.
        """
        fetch_model_weights_path = get_fnames(name="evac_default_tf_weights")
        print(f"fetched {fetch_model_weights_path}")
        self.load_model_weights(fetch_model_weights_path)

    def load_model_weights(self, weights_path):
        """Load the custom pre-training weights to use for the fitting.

        Parameters
        ----------
        weights_path : str
            Path to the file containing the weights (hdf5, saved by tensorflow)
        """
        try:
            self.model.load_weights(weights_path)
        except ValueError as e:
            raise ValueError(
                "Expected input for the provided model weights \
                             do not match the declared model"
            ) from e

    def __predict(self, x_test):
        """Internal prediction function

        Parameters
        ----------
        x_test : np.ndarray (batch, 128, 128, 128, 1)
            Image should match the required shape of the model.

        Returns
        -------
        np.ndarray (batch, ...)
            Predicted brain mask
        """
        return self.model.predict(x_test)

    @deprecated_params(
        "largest_area", new_name="finalize_mask", since="1.10", until="1.12"
    )
    def predict(
        self,
        T1,
        affine,
        *,
        voxsize=(1, 1, 1),
        batch_size=None,
        return_affine=False,
        return_prob=False,
        finalize_mask=True,
    ):
        """Wrapper function to facilitate prediction of larger dataset.

        Parameters
        ----------
        T1 : np.ndarray or list of np.ndarray
            For a single image, input should be a 3D array.
            If multiple images, it should be a a list or tuple.
            or list of np.ndarrays with len of batch_size
        affine : np.ndarray (4, 4) or (batch, 4, 4)
            Affine matrix for the T1 image. Should have
            batch dimension if T1 has one.
        voxsize : np.ndarray or list or tuple, optional
            (3,) or (batch, 3)
            voxel size of the T1 image.
        batch_size : int, optional
            Number of images per prediction pass. Only available if data
            is provided with a batch dimension.
            Consider lowering it if you get an out of memory error.
            Increase it if you want it to be faster and have a lot of data.
            If None, batch_size will be set to 1 if the provided image
            has a batch dimension.
        return_affine : bool, optional
            Whether to return the affine matrix. Useful if the input was a
            file path.
        return_prob : bool, optional
            Whether to return the probability map instead of a
            binary mask. Useful for testing.
        finalize_mask : bool, optional
            Whether to remove potential holes or islands.
            Useful for solving minor errors.

        Returns
        -------
        pred_output : np.ndarray (...) or (batch, ...)
            Predicted brain mask

        affine : np.ndarray (...) or (batch, ...)
            affine matrix of mask
            only if return_affine is True
        """
        voxsize = np.array(voxsize)
        affine = np.array(affine)

        if isinstance(T1, (list, tuple)):
            dim = 4
            T1 = np.array(T1)
        elif len(T1.shape) == 3:
            dim = 3
            if batch_size is not None:
                logger.warning(
                    "Batch size specified, but not used",
                    "due to the input not having \
                                a batch dimension",
                )

            T1 = np.expand_dims(T1, 0)
            affine = np.expand_dims(affine, 0)
            voxsize = np.expand_dims(voxsize, 0)
        else:
            raise ValueError(
                "T1 data should be a np.ndarray of dimension 3 or a list/tuple of it"
            )
        if batch_size is None:
            batch_size = 1

        input_data = np.zeros((128, 128, 128, len(T1)))
        affines = np.zeros((len(T1), 4, 4))
        mid_shapes = np.zeros((len(T1), 3)).astype(int)
        offset_arrays = np.zeros((len(T1), 4, 4)).astype(int)
        scales = np.zeros(len(T1))
        crop_vss = np.zeros((len(T1), 3, 2))
        pad_vss = np.zeros((len(T1), 3, 2))

        # Normalize the data.
        n_T1 = np.zeros(T1.shape)
        for i, T1_img in enumerate(T1):
            n_T1[i] = normalize(T1_img, new_min=0, new_max=1)
            t_img, t_affine, mid_shape, offset_array, scale, crop_vs, pad_vs = (
                transform_img(n_T1[i], affine[i], voxsize=voxsize[i])
            )
            input_data[..., i] = t_img
            affines[i] = t_affine
            mid_shapes[i] = mid_shape
            offset_arrays[i] = offset_array
            scales[i] = scale
            crop_vss[i] = crop_vs
            pad_vss[i] = pad_vs

        # Prediction stage
        prediction = np.zeros((len(T1), 128, 128, 128), dtype=np.float32)
        for batch_idx in range(batch_size, len(T1) + 1, batch_size):
            batch = input_data[..., batch_idx - batch_size : batch_idx]
            temp_input = prepare_img(batch)
            temp_pred = self.__predict(temp_input)
            prediction[:batch_idx] = temp_pred
        remainder = np.mod(len(T1), batch_size)
        if remainder != 0:
            temp_input = prepare_img(input_data[..., -remainder:])
            temp_pred = self.__predict(temp_input)
            prediction[-remainder:] = temp_pred

        output_mask = []
        for i in range(len(T1)):
            output = recover_img(
                prediction[i],
                affines[i],
                mid_shapes[i],
                n_T1[i].shape,
                offset_arrays[i],
                voxsize=voxsize[i],
                scale=scales[i],
                crop_vs=crop_vss[i],
                pad_vs=pad_vss[i],
            )
            if not return_prob:
                output = np.where(output >= 0.5, 1, 0)
                if finalize_mask:
                    output = remove_holes_and_islands(output, slice_wise=True)
            output_mask.append(output)

        if dim == 3:
            output_mask = output_mask[0]
            affine = affine[0]

        output_mask = np.array(output_mask)
        affine = np.array(affine)
        if return_affine:
            return output_mask, affine
        else:
            return output_mask



from functools import partial

import numpy as np

from dipy.testing.decorators import warning_for_keywords
from dipy.viz.horizon.tab import (
    HorizonTab,
    build_checkbox,
    build_label,
    build_radio_button,
    build_slider,
)


class PeaksTab(HorizonTab):
    def __init__(self, peak_actor):
        """Initialize Interaction tab for peaks visualization.

        Parameters
        ----------
        peak_actor : PeaksActor
            Horizon PeaksActor visualize peaks.
        """
        super().__init__()

        self._actor = peak_actor
        self._name = "Peaks"

        self._tab_id = 0

        self._actor_toggle = build_checkbox(
            labels=[""], checked_labels=[""], on_change=self._toggle_actors
        )

        self._opacity_label, self._opacity = build_slider(
            initial_value=1.0,
            max_value=1.0,
            text_template="{ratio:.0%}",
            on_change=self._change_opacity,
            label="Opacity",
        )

        min_centers = self._actor.min_centers
        max_centers = self._actor.max_centers
        cross_section = self._actor.cross_section

        self._slice_x_label, self._slice_x = build_slider(
            initial_value=cross_section[0],
            min_value=min_centers[0],
            max_value=max_centers[0],
            text_template="{value:.0f}",
            label="X Slice",
        )
        self._change_slice_x = partial(self._change_slice, selected_slice=self._slice_x)
        self._adjust_slice_x = partial(self._change_slice_x, sync_slice=True)
        self._slice_x.obj.on_moving_slider = self._change_slice_x
        self._slice_x.obj.on_value_changed = self._adjust_slice_x

        self._slice_y_label, self._slice_y = build_slider(
            initial_value=cross_section[1],
            min_value=min_centers[1],
            max_value=max_centers[1],
            text_template="{value:.0f}",
            label="Y Slice",
        )
        self._change_slice_y = partial(self._change_slice, selected_slice=self._slice_y)
        self._adjust_slice_y = partial(self._change_slice_y, sync_slice=True)
        self._slice_y.obj.on_moving_slider = self._change_slice_y
        self._slice_y.obj.on_value_changed = self._adjust_slice_y

        self._slice_z_label, self._slice_z = build_slider(
            initial_value=cross_section[2],
            min_value=min_centers[2],
            max_value=max_centers[2],
            text_template="{value:.0f}",
            label="Z Slice",
        )
        self._change_slice_z = partial(self._change_slice, selected_slice=self._slice_z)
        self._adjust_slice_z = partial(self._change_slice_z, sync_slice=True)
        self._slice_z.obj.on_moving_slider = self._change_slice_z
        self._slice_z.obj.on_value_changed = self._adjust_slice_z

        low_ranges = self._actor.low_ranges
        high_ranges = self._actor.high_ranges

        self._range_x_label, self._range_x = build_slider(
            initial_value=(low_ranges[0], high_ranges[0]),
            min_value=low_ranges[0],
            max_value=high_ranges[0],
            text_template="{value:.0f}",
            label="X Range",
            is_double_slider=True,
        )
        self._change_range_x = partial(self._change_range, selected_range=self._range_x)
        self._range_x.obj.on_change = self._change_range_x

        self._range_y_label, self._range_y = build_slider(
            initial_value=(low_ranges[1], high_ranges[1]),
            min_value=low_ranges[1],
            max_value=high_ranges[1],
            text_template="{value:.0f}",
            label="Y Range",
            is_double_slider=True,
        )
        self._change_range_y = partial(self._change_range, selected_range=self._range_y)
        self._range_y.obj.on_change = self._change_range_y

        self._range_z_label, self._range_z = build_slider(
            initial_value=(low_ranges[2], high_ranges[2]),
            min_value=low_ranges[2],
            max_value=high_ranges[2],
            text_template="{value:.0f}",
            label="Z Range",
            is_double_slider=True,
        )
        self._change_range_z = partial(self._change_range, selected_range=self._range_z)
        self._range_z.obj.on_change = self._change_range_z

        self._view_mode_label = build_label(text="View Mode")

        self._view_modes = ["Cross section", "Range"]
        self._view_mode_toggler = build_radio_button(
            labels=self._view_modes,
            checked_labels=[self._view_modes[0]],
            padding=1.5,
            on_change=self._toggle_view_mode,
        )

        self._register_elements(
            self._opacity_label,
            self._opacity,
            self._actor_toggle,
            self._slice_x_label,
            self._slice_x,
            self._slice_y_label,
            self._slice_y,
            self._slice_z_label,
            self._slice_z,
            self._range_x_label,
            self._range_x,
            self._range_y_label,
            self._range_y,
            self._range_z_label,
            self._range_z,
            self._view_mode_label,
            self._view_mode_toggler,
        )

    def _change_opacity(self, slider):
        """Update opacity of the peaks actor by adjusting the slider to
        suitable value.

        Parameters
        ----------
        slider : LineSlider2D
            FURY slider UI element.
        """
        self._actor.global_opacity = slider.value

    def _change_range(self, slider, selected_range):
        """Update the range of peaks actor by adjusting the double slider.
        Only usable in Range view mode.

        Parameters
        ----------
        slider : LineDoubleSlider2D
            FURY two side slider UI element.
        selected_range : HorizonUIElement
            Selected horizon ui element intended to change.
        """
        selected_range.selected_value = (
            slider.left_disk_value,
            slider.right_disk_value,
        )
        self._actor.display_extent(
            self._range_x.selected_value[0],
            self._range_x.selected_value[1],
            self._range_y.selected_value[0],
            self._range_y.selected_value[1],
            self._range_z.selected_value[0],
            self._range_z.selected_value[1],
        )

    @warning_for_keywords()
    def _change_slice(self, slider, selected_slice, *, sync_slice=False):
        """Update the slice value of peaks actor by adjusting the slider. Only
        usable in Cross Section view mode.

        Parameters
        ----------
        slider : LineSlider2D
            FURY slider UI element.
        selected_slice : HorizonUIElement
            Selected horizon ui element intended to change.
        sync_slice : bool, optional
            Whether the slice is getting synchronized some other change,
            by default False
        """
        value = int(np.rint(slider.value))
        selected_slice.selected_value = value

        if not sync_slice:
            self.on_slice_change(
                self._tab_id,
                self._slice_x.selected_value,
                self._slice_y.selected_value,
                self._slice_z.selected_value,
            )

        if self._view_mode_toggler.selected_value[0] == self._view_modes[0]:
            self._actor.display_cross_section(
                self._slice_x.selected_value,
                self._slice_y.selected_value,
                self._slice_z.selected_value,
            )

    def _show_cross_section(self):
        """Show Cross Section view mode. Hide the Range sliders and labels."""
        self.hide(
            self._range_x_label,
            self._range_x,
            self._range_y_label,
            self._range_y,
            self._range_z_label,
            self._range_z,
        )
        self.show(
            self._slice_x_label,
            self._slice_x,
            self._slice_y_label,
            self._slice_y,
            self._slice_z_label,
            self._slice_z,
        )
        self._change_slice(self._slice_x.obj, self._slice_x, sync_slice=True)

    def _show_range(self):
        """Show Range view mode. Hide Cross Section sliders and labels."""
        self.hide(
            self._slice_x_label,
            self._slice_x,
            self._slice_y_label,
            self._slice_y,
            self._slice_z_label,
            self._slice_z,
        )
        self.show(
            self._range_x_label,
            self._range_x,
            self._range_y_label,
            self._range_y,
            self._range_z_label,
            self._range_z,
        )

    def _toggle_view_mode(self, radio):
        self._view_mode_toggler.selected_value = radio.checked_labels
        if radio.checked_labels[0] == self._view_modes[0]:
            self._actor.display_cross_section(
                self._actor.cross_section[0],
                self._actor.cross_section[1],
                self._actor.cross_section[2],
            )
            self._show_cross_section()
        else:
            self._actor.display_extent(
                self._actor.low_ranges[0],
                self._actor.high_ranges[0],
                self._actor.low_ranges[1],
                self._actor.high_ranges[1],
                self._actor.low_ranges[2],
                self._actor.high_ranges[2],
            )
            self._show_range()

    def on_tab_selected(self):
        """Trigger when tab becomes active."""
        super().on_tab_selected()
        self._toggle_view_mode(self._view_mode_toggler.obj)

    def update_slices(self, x_slice, y_slice, z_slice):
        """Updates slicer positions.

        Parameters
        ----------
        x_slice: float
            x-value where the slicer should be placed
        y_slice: float
            y-value where the slicer should be placed
        z_slice: float
            z-value where the slicer should be placed
        """
        if not self._slice_x.obj.value == x_slice:
            self._slice_x.obj.value = x_slice

        if not self._slice_y.obj.value == y_slice:
            self._slice_y.obj.value = y_slice

        if not self._slice_z.obj.value == z_slice:
            self._slice_z.obj.value = z_slice

    def build(self, tab_id):
        """Build all the elements under the tab.

        Parameters
        ----------
        tab_id : int
            Id of the tab.
        """
        self._tab_id = tab_id

        x_pos = 0.02
        self._actor_toggle.position = (x_pos, 0.85)

        x_pos = 0.04
        self._opacity_label.position = (x_pos, 0.85)
        self._slice_x_label.position = (x_pos, 0.62)
        self._slice_y_label.position = (x_pos, 0.38)
        self._slice_z_label.position = (x_pos, 0.15)
        self._range_x_label.position = (x_pos, 0.62)
        self._range_y_label.position = (x_pos, 0.38)
        self._range_z_label.position = (x_pos, 0.15)

        x_pos = 0.10
        self._opacity.position = (x_pos, 0.85)
        self._slice_x.position = (x_pos, 0.62)
        self._slice_y.position = (x_pos, 0.38)
        self._slice_z.position = (x_pos, 0.15)

        x_pos = 0.105
        self._range_x.position = (x_pos, 0.66)
        self._range_y.position = (x_pos, 0.42)
        self._range_z.position = (x_pos, 0.19)

        x_pos = 0.52
        self._view_mode_label.position = (x_pos, 0.85)

        x_pos = 0.62
        self._view_mode_toggler.position = (x_pos, 0.80)

        cross_section = self._actor.cross_section
        self._actor.display_cross_section(
            cross_section[0], cross_section[1], cross_section[2]
        )

    @property
    def name(self):
        """Name of the tab.

        Returns
        -------
        str
        """
        return self._name

    @property
    def actors(self):
        """actors controlled by tab.

        Returns
        -------
        list
            List of actors.
        """
        return [self._actor]
